# Generated by Django 2.1.4 on 2019-02-27 21:37

import django.core.validators
import django.db.models.deletion
from django.db import migrations, models

import djstripe.enums
import djstripe.fields


def postgres_migration_prep(apps, schema_editor):
    """
    Set null text fields to empty string to workaround incompatibility with migration 0003 on postgres
    See https://github.com/dj-stripe/dj-stripe/issues/850
    """

    Account = apps.get_model("djstripe", "Account")
    BankAccount = apps.get_model("djstripe", "BankAccount")
    Card = apps.get_model("djstripe", "Card")
    Charge = apps.get_model("djstripe", "Charge")
    Customer = apps.get_model("djstripe", "Customer")
    Event = apps.get_model("djstripe", "Event")
    Invoice = apps.get_model("djstripe", "Invoice")
    Payout = apps.get_model("djstripe", "Payout")
    Plan = apps.get_model("djstripe", "Plan")
    Product = apps.get_model("djstripe", "Product")
    Refund = apps.get_model("djstripe", "Refund")
    Source = apps.get_model("djstripe", "Source")
    Transfer = apps.get_model("djstripe", "Transfer")

    model_fields = [
        (
            Account,
            (
                "business_name",
                "business_primary_color",
                "business_url",
                "payout_statement_descriptor",
                "product_description",
                "support_url",
            ),
        ),
        (BankAccount, ("account_holder_name",)),
        (
            Card,
            (
                "address_city",
                "address_country",
                "address_line1",
                "address_line1_check",
                "address_line2",
                "address_state",
                "address_zip",
                "address_zip_check",
                "country",
                "cvc_check",
                "dynamic_last4",
                "fingerprint",
                "name",
                "tokenization_method",
            ),
        ),
        (
            Charge,
            (
                "failure_code",
                "failure_message",
                "receipt_email",
                "receipt_number",
                "statement_descriptor",
                "transfer_group",
            ),
        ),
        (Customer, ("business_vat_id", "currency", "email")),
        (Event, ("idempotency_key", "request_id")),
        (
            Invoice,
            ("hosted_invoice_url", "invoice_pdf", "number", "statement_descriptor"),
        ),
        (Payout, ("failure_code", "failure_message", "statement_descriptor")),
        (Plan, ("aggregate_usage", "billing_scheme", "nickname")),
        (Product, ("caption", "statement_descriptor", "unit_label")),
        (Refund, ("failure_reason", "reason", "receipt_number")),
        (Source, ("currency", "statement_descriptor")),
        (Transfer, ("transfer_group",)),
    ]

    for model, fields in model_fields:
        for field in fields:
            filter_param = {"{}__isnull".format(field): True}
            update_param = {field: ""}
            model.objects.filter(**filter_param).update(**update_param)


class Migration(migrations.Migration):
    def __init__(self, *args, **kwargs):
        # Hack to support sqlite for quick testing. Without this migrations fail with:
        #    "django.db.utils.NotSupportedError: Renaming the 'djstripe_paymentmethod' table while in a transaction
        #    is not supported on SQLite because it would break referential integrity.
        #    Try adding `atomic = False` to the Migration class."
        from django.db import connection

        # use getattr because I think connection.vendor isn't documented
        if getattr(connection, "vendor", None) == "sqlite":
            self.atomic = False

        super().__init__(*args, **kwargs)

    dependencies = [("djstripe", "0002_auto_20180627_1121")]

    operations = [
        migrations.RunPython(postgres_migration_prep, migrations.RunPython.noop),
        migrations.AlterField(
            model_name="account",
            name="display_name",
            field=models.CharField(
                blank=True,
                default="",
                help_text="The display name for this account. This is used on the Stripe Dashboard to differentiate between accounts.",
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="account",
            name="support_email",
            field=models.CharField(
                blank=True,
                default="",
                help_text="A publicly shareable support email address for the business",
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="account",
            name="support_phone",
            field=models.CharField(
                blank=True,
                default="",
                help_text="A publicly shareable support phone number for the business",
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="card",
            name="customer",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="legacy_cards",
                to="djstripe.Customer",
            ),
        ),
        migrations.AddField(
            model_name="coupon",
            name="name",
            field=models.TextField(
                blank=True,
                default="",
                help_text="Name of the coupon displayed to customers on for instance invoices or receipts.",
                max_length=5000,
            ),
        ),
        migrations.AlterField(
            model_name="source",
            name="type",
            field=djstripe.fields.StripeEnumField(
                enum=djstripe.enums.SourceType,
                help_text="The type of the source.",
                max_length=20,
            ),
        ),
        migrations.AddField(
            model_name="plan",
            name="active",
            field=models.BooleanField(
                default=True,
                help_text="Whether the plan is currently available for new subscriptions.",
            ),
            preserve_default=False,
        ),
        migrations.RemoveField(model_name="charge", name="fraudulent"),
        migrations.RemoveField(model_name="charge", name="receipt_sent"),
        migrations.RemoveField(model_name="charge", name="source_stripe_id"),
        migrations.RemoveField(model_name="charge", name="source_type"),
        migrations.RemoveField(model_name="charge", name="fee"),
        migrations.RemoveField(model_name="charge", name="fee_details"),
        migrations.RemoveField(model_name="transfer", name="date"),
        migrations.RemoveField(model_name="transfer", name="destination_type"),
        migrations.RemoveField(model_name="transfer", name="failure_code"),
        migrations.RemoveField(model_name="transfer", name="failure_message"),
        migrations.RemoveField(model_name="transfer", name="fee"),
        migrations.RemoveField(model_name="transfer", name="fee_details"),
        migrations.RemoveField(model_name="transfer", name="statement_descriptor"),
        migrations.RemoveField(model_name="transfer", name="status"),
        migrations.AlterModelOptions(
            name="account", options={"get_latest_by": "created"}
        ),
        migrations.AlterModelOptions(
            name="bankaccount", options={"get_latest_by": "created"}
        ),
        migrations.AlterModelOptions(name="card", options={"get_latest_by": "created"}),
        migrations.AlterModelOptions(
            name="charge", options={"get_latest_by": "created"}
        ),
        migrations.AlterModelOptions(
            name="dispute", options={"get_latest_by": "created"}
        ),
        migrations.AlterModelOptions(
            name="event", options={"get_latest_by": "created"}
        ),
        migrations.AlterModelOptions(
            name="fileupload", options={"get_latest_by": "created"}
        ),
        migrations.AlterModelOptions(
            name="invoiceitem", options={"get_latest_by": "created"}
        ),
        migrations.AlterModelOptions(
            name="payout", options={"get_latest_by": "created"}
        ),
        migrations.AlterModelOptions(
            name="product", options={"get_latest_by": "created"}
        ),
        migrations.AlterModelOptions(
            name="refund", options={"get_latest_by": "created"}
        ),
        migrations.AlterModelOptions(
            name="source", options={"get_latest_by": "created"}
        ),
        migrations.AlterModelOptions(
            name="subscription", options={"get_latest_by": "created"}
        ),
        migrations.AlterModelOptions(
            name="transfer", options={"get_latest_by": "created"}
        ),
        migrations.AlterModelOptions(
            name="upcominginvoice", options={"get_latest_by": "created"}
        ),
        migrations.CreateModel(
            name="CountrySpec",
            fields=[
                ("djstripe_created", models.DateTimeField(auto_now_add=True)),
                ("djstripe_updated", models.DateTimeField(auto_now=True)),
                (
                    "id",
                    models.CharField(max_length=2, primary_key=True, serialize=False),
                ),
                (
                    "default_currency",
                    djstripe.fields.StripeCurrencyCodeField(
                        help_text="The default currency for this country. This applies to both payment methods and bank accounts.",
                        max_length=3,
                    ),
                ),
                (
                    "supported_bank_account_currencies",
                    djstripe.fields.JSONField(
                        help_text="Currencies that can be accepted in the specific country (for transfers)."
                    ),
                ),
                (
                    "supported_payment_currencies",
                    djstripe.fields.JSONField(
                        help_text="Currencies that can be accepted in the specified country (for payments)."
                    ),
                ),
                (
                    "supported_payment_methods",
                    djstripe.fields.JSONField(
                        help_text="Payment methods available in the specified country."
                    ),
                ),
                (
                    "supported_transfer_countries",
                    djstripe.fields.JSONField(
                        help_text="Countries that can accept transfers from the specified country."
                    ),
                ),
                (
                    "verification_fields",
                    djstripe.fields.JSONField(
                        help_text="Lists the types of verification data needed to keep an account open."
                    ),
                ),
            ],
            options={"abstract": False},
        ),
        migrations.CreateModel(
            name="BalanceTransaction",
            fields=[
                (
                    "djstripe_id",
                    models.BigAutoField(
                        primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("id", djstripe.fields.StripeIdField(max_length=255, unique=True)),
                (
                    "livemode",
                    models.NullBooleanField(
                        default=None,
                        help_text="Null here indicates that the livemode status is unknown or was previously unrecorded. Otherwise, this field indicates whether this record comes from Stripe test mode or live mode operation.",
                    ),
                ),
                (
                    "created",
                    djstripe.fields.StripeDateTimeField(
                        blank=True,
                        help_text="The datetime this object was created in stripe.",
                        null=True,
                    ),
                ),
                (
                    "metadata",
                    djstripe.fields.JSONField(
                        blank=True,
                        help_text="A set of key/value pairs that you can attach to an object. It can be useful for storing additional information about an object in a structured format.",
                        null=True,
                    ),
                ),
                (
                    "description",
                    models.TextField(
                        blank=True, help_text="A description of this object.", null=True
                    ),
                ),
                ("djstripe_created", models.DateTimeField(auto_now_add=True)),
                ("djstripe_updated", models.DateTimeField(auto_now=True)),
                (
                    "amount",
                    djstripe.fields.StripeQuantumCurrencyAmountField(
                        help_text="Gross amount of the transaction, in cents."
                    ),
                ),
                (
                    "available_on",
                    djstripe.fields.StripeDateTimeField(
                        help_text="The date the transaction's net funds will become available in the Stripe balance."
                    ),
                ),
                (
                    "currency",
                    djstripe.fields.StripeCurrencyCodeField(
                        help_text="Three-letter ISO currency code", max_length=3
                    ),
                ),
                (
                    "exchange_rate",
                    models.DecimalField(null=True, decimal_places=6, max_digits=8),
                ),
                (
                    "fee",
                    djstripe.fields.StripeQuantumCurrencyAmountField(
                        help_text="Fee (in cents) paid for this transaction."
                    ),
                ),
                ("fee_details", djstripe.fields.JSONField()),
                (
                    "net",
                    djstripe.fields.StripeQuantumCurrencyAmountField(
                        help_text="Net amount of the transaction, in cents."
                    ),
                ),
                (
                    "status",
                    djstripe.fields.StripeEnumField(
                        enum=djstripe.enums.BalanceTransactionStatus, max_length=9
                    ),
                ),
                (
                    "type",
                    djstripe.fields.StripeEnumField(
                        enum=djstripe.enums.BalanceTransactionType, max_length=22
                    ),
                ),
            ],
            options={"get_latest_by": "created", "abstract": False},
        ),
        migrations.CreateModel(
            name="ScheduledQueryRun",
            fields=[
                (
                    "djstripe_id",
                    models.BigAutoField(
                        primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("id", djstripe.fields.StripeIdField(max_length=255, unique=True)),
                (
                    "livemode",
                    models.NullBooleanField(
                        default=None,
                        help_text="Null here indicates that the livemode status is unknown or was previously unrecorded. Otherwise, this field indicates whether this record comes from Stripe test mode or live mode operation.",
                    ),
                ),
                (
                    "created",
                    djstripe.fields.StripeDateTimeField(
                        blank=True,
                        help_text="The datetime this object was created in stripe.",
                        null=True,
                    ),
                ),
                (
                    "metadata",
                    djstripe.fields.JSONField(
                        blank=True,
                        help_text="A set of key/value pairs that you can attach to an object. It can be useful for storing additional information about an object in a structured format.",
                        null=True,
                    ),
                ),
                (
                    "description",
                    models.TextField(
                        blank=True, help_text="A description of this object.", null=True
                    ),
                ),
                ("djstripe_created", models.DateTimeField(auto_now_add=True)),
                ("djstripe_updated", models.DateTimeField(auto_now=True)),
                (
                    "data_load_time",
                    djstripe.fields.StripeDateTimeField(
                        help_text="When the query was run, Sigma contained a snapshot of your Stripe data at this time."
                    ),
                ),
                (
                    "error",
                    djstripe.fields.JSONField(
                        blank=True,
                        help_text="If the query run was not succeesful, contains information about the failure.",
                        null=True,
                    ),
                ),
                (
                    "result_available_until",
                    djstripe.fields.StripeDateTimeField(
                        help_text="Time at which the result expires and is no longer available for download."
                    ),
                ),
                (
                    "sql",
                    models.TextField(help_text="SQL for the query.", max_length=5000),
                ),
                (
                    "status",
                    djstripe.fields.StripeEnumField(
                        enum=djstripe.enums.ScheduledQueryRunStatus,
                        help_text="The query's execution status.",
                        max_length=9,
                    ),
                ),
                (
                    "title",
                    models.TextField(help_text="Title of the query.", max_length=5000),
                ),
                (
                    "file",
                    models.ForeignKey(
                        blank=True,
                        help_text="The file object representing the results of the query.",
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="djstripe.FileUpload",
                    ),
                ),
            ],
            options={"get_latest_by": "created", "abstract": False},
        ),
        migrations.CreateModel(
            name="SubscriptionItem",
            fields=[
                (
                    "djstripe_id",
                    models.BigAutoField(
                        primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("id", djstripe.fields.StripeIdField(max_length=255, unique=True)),
                (
                    "livemode",
                    models.NullBooleanField(
                        default=None,
                        help_text="Null here indicates that the livemode status is unknown or was previously unrecorded. Otherwise, this field indicates whether this record comes from Stripe test mode or live mode operation.",
                    ),
                ),
                (
                    "created",
                    djstripe.fields.StripeDateTimeField(
                        blank=True,
                        help_text="The datetime this object was created in stripe.",
                        null=True,
                    ),
                ),
                (
                    "metadata",
                    djstripe.fields.JSONField(
                        blank=True,
                        help_text="A set of key/value pairs that you can attach to an object. It can be useful for storing additional information about an object in a structured format.",
                        null=True,
                    ),
                ),
                (
                    "description",
                    models.TextField(
                        blank=True, help_text="A description of this object.", null=True
                    ),
                ),
                ("djstripe_created", models.DateTimeField(auto_now_add=True)),
                ("djstripe_updated", models.DateTimeField(auto_now=True)),
                (
                    "quantity",
                    models.PositiveIntegerField(
                        help_text="The quantity of the plan to which the customer should be subscribed."
                    ),
                ),
                (
                    "plan",
                    models.ForeignKey(
                        help_text="The plan the customer is subscribed to.",
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="subscription_items",
                        to="djstripe.Plan",
                    ),
                ),
                (
                    "subscription",
                    models.ForeignKey(
                        help_text="The subscription this subscription item belongs to.",
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="items",
                        to="djstripe.Subscription",
                    ),
                ),
            ],
            options={"get_latest_by": "created", "abstract": False},
        ),
        migrations.CreateModel(
            name="TransferReversal",
            fields=[
                (
                    "djstripe_id",
                    models.BigAutoField(
                        primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("id", djstripe.fields.StripeIdField(max_length=255, unique=True)),
                (
                    "livemode",
                    models.NullBooleanField(
                        default=None,
                        help_text="Null here indicates that the livemode status is unknown or was previously unrecorded. Otherwise, this field indicates whether this record comes from Stripe test mode or live mode operation.",
                    ),
                ),
                (
                    "created",
                    djstripe.fields.StripeDateTimeField(
                        blank=True,
                        help_text="The datetime this object was created in stripe.",
                        null=True,
                    ),
                ),
                (
                    "metadata",
                    djstripe.fields.JSONField(
                        blank=True,
                        help_text="A set of key/value pairs that you can attach to an object. It can be useful for storing additional information about an object in a structured format.",
                        null=True,
                    ),
                ),
                (
                    "description",
                    models.TextField(
                        blank=True, help_text="A description of this object.", null=True
                    ),
                ),
                ("djstripe_created", models.DateTimeField(auto_now_add=True)),
                ("djstripe_updated", models.DateTimeField(auto_now=True)),
                (
                    "amount",
                    djstripe.fields.StripeQuantumCurrencyAmountField(
                        help_text="Amount, in cents."
                    ),
                ),
                (
                    "currency",
                    djstripe.fields.StripeCurrencyCodeField(
                        help_text="Three-letter ISO currency code", max_length=3
                    ),
                ),
                (
                    "balance_transaction",
                    models.ForeignKey(
                        blank=True,
                        help_text="Balance transaction that describes the impact on your account balance.",
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="transfer_reversals",
                        to="djstripe.BalanceTransaction",
                    ),
                ),
                (
                    "transfer",
                    models.ForeignKey(
                        help_text="The transfer that was reversed.",
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="reversals",
                        to="djstripe.Transfer",
                    ),
                ),
            ],
            options={"get_latest_by": "created", "abstract": False},
        ),
        migrations.CreateModel(
            name="UsageRecord",
            fields=[
                (
                    "djstripe_id",
                    models.BigAutoField(
                        primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("id", djstripe.fields.StripeIdField(max_length=255, unique=True)),
                (
                    "livemode",
                    models.NullBooleanField(
                        default=None,
                        help_text="Null here indicates that the livemode status is unknown or was previously unrecorded. Otherwise, this field indicates whether this record comes from Stripe test mode or live mode operation.",
                    ),
                ),
                (
                    "created",
                    djstripe.fields.StripeDateTimeField(
                        blank=True,
                        help_text="The datetime this object was created in stripe.",
                        null=True,
                    ),
                ),
                (
                    "metadata",
                    djstripe.fields.JSONField(
                        blank=True,
                        help_text="A set of key/value pairs that you can attach to an object. It can be useful for storing additional information about an object in a structured format.",
                        null=True,
                    ),
                ),
                (
                    "description",
                    models.TextField(
                        blank=True, help_text="A description of this object.", null=True
                    ),
                ),
                ("djstripe_created", models.DateTimeField(auto_now_add=True)),
                ("djstripe_updated", models.DateTimeField(auto_now=True)),
                (
                    "quantity",
                    models.PositiveIntegerField(
                        help_text="The quantity of the plan to which the customer should be subscribed."
                    ),
                ),
                (
                    "subscription_item",
                    models.ForeignKey(
                        help_text="The subscription item this usage record contains data for.",
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="usage_records",
                        to="djstripe.SubscriptionItem",
                    ),
                ),
            ],
            options={"get_latest_by": "created", "abstract": False},
        ),
        migrations.CreateModel(
            name="ApplicationFee",
            fields=[
                (
                    "djstripe_id",
                    models.BigAutoField(
                        primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("id", djstripe.fields.StripeIdField(max_length=255, unique=True)),
                (
                    "livemode",
                    models.NullBooleanField(
                        default=None,
                        help_text="Null here indicates that the livemode status is unknown or was previously unrecorded. Otherwise, this field indicates whether this record comes from Stripe test mode or live mode operation.",
                    ),
                ),
                (
                    "created",
                    djstripe.fields.StripeDateTimeField(
                        blank=True,
                        help_text="The datetime this object was created in stripe.",
                        null=True,
                    ),
                ),
                (
                    "metadata",
                    djstripe.fields.JSONField(
                        blank=True,
                        help_text="A set of key/value pairs that you can attach to an object. It can be useful for storing additional information about an object in a structured format.",
                        null=True,
                    ),
                ),
                (
                    "description",
                    models.TextField(
                        blank=True, help_text="A description of this object.", null=True
                    ),
                ),
                ("djstripe_created", models.DateTimeField(auto_now_add=True)),
                ("djstripe_updated", models.DateTimeField(auto_now=True)),
                (
                    "amount",
                    djstripe.fields.StripeQuantumCurrencyAmountField(
                        help_text="Amount earned, in cents."
                    ),
                ),
                (
                    "amount_refunded",
                    djstripe.fields.StripeQuantumCurrencyAmountField(
                        help_text="Amount in cents refunded (can be less than the amount attribute on the fee if a partial refund was issued)"
                    ),
                ),
                (
                    "currency",
                    djstripe.fields.StripeCurrencyCodeField(
                        help_text="Three-letter ISO currency code", max_length=3
                    ),
                ),
                (
                    "refunded",
                    models.BooleanField(
                        help_text="Whether the fee has been fully refunded. If the fee is only partially refunded, this attribute will still be false."
                    ),
                ),
                (
                    "balance_transaction",
                    models.ForeignKey(
                        help_text="Balance transaction that describes the impact on your account balance.",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="djstripe.BalanceTransaction",
                    ),
                ),
                (
                    "charge",
                    models.ForeignKey(
                        help_text="The charge that the application fee was taken from.",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="djstripe.Charge",
                    ),
                ),
            ],
            options={"get_latest_by": "created", "abstract": False},
        ),
        migrations.CreateModel(
            name="ApplicationFeeRefund",
            fields=[
                (
                    "djstripe_id",
                    models.BigAutoField(
                        primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("id", djstripe.fields.StripeIdField(max_length=255, unique=True)),
                (
                    "livemode",
                    models.NullBooleanField(
                        default=None,
                        help_text="Null here indicates that the livemode status is unknown or was previously unrecorded. Otherwise, this field indicates whether this record comes from Stripe test mode or live mode operation.",
                    ),
                ),
                (
                    "created",
                    djstripe.fields.StripeDateTimeField(
                        blank=True,
                        help_text="The datetime this object was created in stripe.",
                        null=True,
                    ),
                ),
                (
                    "metadata",
                    djstripe.fields.JSONField(
                        blank=True,
                        help_text="A set of key/value pairs that you can attach to an object. It can be useful for storing additional information about an object in a structured format.",
                        null=True,
                    ),
                ),
                ("djstripe_created", models.DateTimeField(auto_now_add=True)),
                ("djstripe_updated", models.DateTimeField(auto_now=True)),
                (
                    "amount",
                    djstripe.fields.StripeQuantumCurrencyAmountField(
                        help_text="Amount refunded, in cents."
                    ),
                ),
                (
                    "currency",
                    djstripe.fields.StripeCurrencyCodeField(
                        help_text="Three-letter ISO currency code", max_length=3
                    ),
                ),
                (
                    "balance_transaction",
                    models.ForeignKey(
                        help_text="Balance transaction that describes the impact on your account balance.",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="djstripe.BalanceTransaction",
                    ),
                ),
                (
                    "fee",
                    models.ForeignKey(
                        help_text="The application fee that was refunded",
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="refunds",
                        to="djstripe.ApplicationFee",
                    ),
                ),
            ],
            options={"get_latest_by": "created", "abstract": False},
        ),
        migrations.AddField(
            model_name="charge",
            name="balance_transaction",
            field=models.ForeignKey(
                help_text="The balance transaction that describes the impact of this charge on your account balance (not including refunds or disputes).",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="djstripe.BalanceTransaction",
            ),
        ),
        migrations.AddField(
            model_name="payout",
            name="balance_transaction",
            field=models.ForeignKey(
                help_text="Balance transaction that describes the impact on your account balance.",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="djstripe.BalanceTransaction",
            ),
        ),
        migrations.AddField(
            model_name="payout",
            name="failure_balance_transaction",
            field=models.ForeignKey(
                help_text="If the payout failed or was canceled, this will be the balance transaction that reversed the initial balance transaction, and puts the funds from the failed payout back in your balance.",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="djstripe.BalanceTransaction",
                related_name="failure_payouts",
            ),
        ),
        migrations.AddField(
            model_name="refund",
            name="balance_transaction",
            field=models.ForeignKey(
                help_text="Balance transaction that describes the impact on your account balance.",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="djstripe.BalanceTransaction",
            ),
        ),
        migrations.AddField(
            model_name="refund",
            name="failure_balance_transaction",
            field=models.ForeignKey(
                help_text="If the refund failed, this balance transaction describes the adjustment made on your account balance that reverses the initial balance transaction.",
                null=True,
                blank=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="djstripe.BalanceTransaction",
                related_name="failure_refunds",
            ),
        ),
        migrations.AddField(
            model_name="transfer",
            name="balance_transaction",
            field=models.ForeignKey(
                blank=True,
                help_text="Balance transaction that describes the impact on your account balance.",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="djstripe.BalanceTransaction",
            ),
        ),
        migrations.RenameField(
            model_name="account", old_name="stripe_id", new_name="id"
        ),
        migrations.RenameField(
            model_name="bankaccount", old_name="stripe_id", new_name="id"
        ),
        migrations.RenameField(model_name="card", old_name="stripe_id", new_name="id"),
        migrations.RenameField(
            model_name="charge", old_name="stripe_id", new_name="id"
        ),
        migrations.RenameField(
            model_name="customer", old_name="stripe_id", new_name="id"
        ),
        migrations.RenameField(
            model_name="dispute", old_name="stripe_id", new_name="id"
        ),
        migrations.RenameField(model_name="event", old_name="stripe_id", new_name="id"),
        migrations.RenameField(
            model_name="fileupload", old_name="stripe_id", new_name="id"
        ),
        migrations.RenameField(
            model_name="invoice", old_name="stripe_id", new_name="id"
        ),
        migrations.RenameField(
            model_name="invoiceitem", old_name="stripe_id", new_name="id"
        ),
        migrations.RenameField(
            model_name="payout", old_name="stripe_id", new_name="id"
        ),
        migrations.RenameField(model_name="plan", old_name="stripe_id", new_name="id"),
        migrations.RenameField(
            model_name="product", old_name="stripe_id", new_name="id"
        ),
        migrations.RenameField(
            model_name="refund", old_name="stripe_id", new_name="id"
        ),
        migrations.RenameField(
            model_name="source", old_name="stripe_id", new_name="id"
        ),
        migrations.RenameField(
            model_name="subscription", old_name="stripe_id", new_name="id"
        ),
        migrations.RenameField(
            model_name="transfer", old_name="stripe_id", new_name="id"
        ),
        migrations.RenameField(
            model_name="coupon", old_name="stripe_id", new_name="id"
        ),
        migrations.AlterField(
            model_name="coupon",
            name="percent_off",
            field=djstripe.fields.StripePercentField(
                blank=True,
                decimal_places=2,
                help_text="Percent that will be taken off the subtotal of any invoices for this customer for the duration of the coupon. For example, a coupon with percent_off of 50 will make a $100 invoice $50 instead.",
                max_digits=5,
                null=True,
                validators=[
                    django.core.validators.MinValueValidator(1),
                    django.core.validators.MaxValueValidator(100),
                ],
            ),
        ),
        # Update all text-type fields to non-null CharField blank=True default=""
        migrations.AlterField(
            model_name="account",
            name="business_name",
            field=models.CharField(
                blank=True,
                default="",
                help_text="The publicly visible name of the business",
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="account",
            name="business_primary_color",
            field=models.CharField(
                blank=True,
                default="",
                help_text="A CSS hex color value representing the primary branding color for this account",
                max_length=7,
            ),
        ),
        migrations.AlterField(
            model_name="account",
            name="business_url",
            field=models.CharField(
                blank=True,
                default="",
                help_text="The publicly visible website of the business",
                max_length=200,
            ),
        ),
        migrations.AlterField(
            model_name="account",
            name="payout_statement_descriptor",
            field=models.CharField(
                blank=True,
                default="",
                help_text="The text that appears on the bank account statement for payouts.",
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="account",
            name="product_description",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Internal-only description of the product sold or service provided by the business. It’s used by Stripe for risk and underwriting purposes.",
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="account",
            name="support_url",
            field=models.CharField(
                blank=True,
                default="",
                help_text="A publicly shareable URL that provides support for this account",
                max_length=200,
            ),
        ),
        migrations.AlterField(
            model_name="bankaccount",
            name="account_holder_name",
            field=models.TextField(
                blank=True,
                default="",
                help_text="The name of the person or business that owns the bank account.",
                max_length=5000,
            ),
        ),
        migrations.AlterField(
            model_name="card",
            name="address_city",
            field=models.TextField(
                blank=True,
                default="",
                help_text="City/District/Suburb/Town/Village.",
                max_length=5000,
            ),
        ),
        migrations.AlterField(
            model_name="card",
            name="address_country",
            field=models.TextField(
                blank=True,
                default="",
                help_text="Billing address country.",
                max_length=5000,
            ),
        ),
        migrations.AlterField(
            model_name="card",
            name="address_line1",
            field=models.TextField(
                blank=True,
                default="",
                help_text="Street address/PO Box/Company name.",
                max_length=5000,
            ),
        ),
        migrations.AlterField(
            model_name="card",
            name="address_line1_check",
            field=djstripe.fields.StripeEnumField(
                blank=True,
                default="",
                enum=djstripe.enums.CardCheckResult,
                help_text="If `address_line1` was provided, results of the check.",
                max_length=11,
            ),
        ),
        migrations.AlterField(
            model_name="card",
            name="address_line2",
            field=models.TextField(
                blank=True,
                default="",
                help_text="Apartment/Suite/Unit/Building.",
                max_length=5000,
            ),
        ),
        migrations.AlterField(
            model_name="card",
            name="address_state",
            field=models.TextField(
                blank=True,
                default="",
                help_text="State/County/Province/Region.",
                max_length=5000,
            ),
        ),
        migrations.AlterField(
            model_name="card",
            name="address_zip",
            field=models.TextField(
                blank=True, default="", help_text="ZIP or postal code.", max_length=5000
            ),
        ),
        migrations.AlterField(
            model_name="card",
            name="address_zip_check",
            field=djstripe.fields.StripeEnumField(
                blank=True,
                default="",
                enum=djstripe.enums.CardCheckResult,
                help_text="If `address_zip` was provided, results of the check.",
                max_length=11,
            ),
        ),
        migrations.AlterField(
            model_name="card",
            name="country",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Two-letter ISO code representing the country of the card.",
                max_length=2,
            ),
        ),
        migrations.AlterField(
            model_name="card",
            name="cvc_check",
            field=djstripe.fields.StripeEnumField(
                blank=True,
                default="",
                enum=djstripe.enums.CardCheckResult,
                help_text="If a CVC was provided, results of the check.",
                max_length=11,
            ),
        ),
        migrations.AlterField(
            model_name="card",
            name="dynamic_last4",
            field=models.CharField(
                blank=True,
                default="",
                help_text="(For tokenized numbers only.) The last four digits of the device account number.",
                max_length=4,
            ),
        ),
        migrations.AlterField(
            model_name="card",
            name="fingerprint",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Uniquely identifies this particular card number.",
                max_length=16,
            ),
        ),
        migrations.AlterField(
            model_name="card",
            name="name",
            field=models.TextField(
                blank=True, default="", help_text="Cardholder name.", max_length=5000
            ),
        ),
        migrations.AlterField(
            model_name="card",
            name="tokenization_method",
            field=djstripe.fields.StripeEnumField(
                blank=True,
                default="",
                enum=djstripe.enums.CardTokenizationMethod,
                help_text="If the card number is tokenized, this is the method that was used.",
                max_length=11,
            ),
        ),
        migrations.AlterField(
            model_name="charge",
            name="failure_code",
            field=djstripe.fields.StripeEnumField(
                blank=True,
                default="",
                enum=djstripe.enums.ApiErrorCode,
                help_text="Error code explaining reason for charge failure if available.",
                max_length=42,
            ),
        ),
        migrations.AlterField(
            model_name="charge",
            name="failure_message",
            field=models.TextField(
                blank=True,
                default="",
                help_text="Message to user further explaining reason for charge failure if available.",
                max_length=5000,
            ),
        ),
        migrations.AlterField(
            model_name="charge",
            name="receipt_email",
            field=models.TextField(
                blank=True,
                default="",
                help_text="The email address that the receipt for this charge was sent to.",
                max_length=800,
            ),
        ),
        migrations.AlterField(
            model_name="charge",
            name="receipt_number",
            field=models.CharField(
                blank=True,
                default="",
                help_text="The transaction number that appears on email receipts sent for this charge.",
                max_length=14,
            ),
        ),
        migrations.AlterField(
            model_name="charge",
            name="statement_descriptor",
            field=models.CharField(
                blank=True,
                default="",
                help_text="An arbitrary string to be displayed on your customer's credit card statement. The statement description may not include <>\"' characters, and will appear on your customer's statement in capital letters. Non-ASCII characters are automatically stripped. While most banks display this information consistently, some may display it incorrectly or not at all.",
                max_length=22,
            ),
        ),
        migrations.AlterField(
            model_name="charge",
            name="transfer_group",
            field=models.CharField(
                blank=True,
                default="",
                help_text="A string that identifies this transaction as part of a group.",
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="customer",
            name="business_vat_id",
            field=models.CharField(
                blank=True,
                default="",
                help_text="The customer's VAT identification number.",
                max_length=20,
            ),
        ),
        migrations.AlterField(
            model_name="customer",
            name="currency",
            field=djstripe.fields.StripeCurrencyCodeField(
                default="",
                help_text="The currency the customer can be charged in for recurring billing purposes",
                max_length=3,
            ),
        ),
        migrations.AlterField(
            model_name="customer",
            name="email",
            field=models.TextField(blank=True, default="", max_length=5000),
        ),
        migrations.AlterField(
            model_name="event",
            name="idempotency_key",
            field=models.TextField(blank=True, default=""),
        ),
        migrations.AlterField(
            model_name="event",
            name="request_id",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Information about the request that triggered this event, for traceability purposes. If empty string then this is an old entry without that data. If Null then this is not an old entry, but a Stripe 'automated' event with no associated request.",
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="invoice",
            name="hosted_invoice_url",
            field=models.TextField(
                blank=True,
                default="",
                help_text="The URL for the hosted invoice page, which allows customers to view and pay an invoice. If the invoice has not been frozen yet, this will be null.",
                max_length=799,
            ),
        ),
        migrations.AlterField(
            model_name="invoice",
            name="invoice_pdf",
            field=models.TextField(
                blank=True,
                default="",
                help_text="The link to download the PDF for the invoice. If the invoice has not been frozen yet, this will be null.",
                max_length=799,
            ),
        ),
        migrations.AlterField(
            model_name="invoice",
            name="number",
            field=models.CharField(
                blank=True,
                default="",
                help_text="A unique, identifying string that appears on emails sent to the customer for this invoice. This starts with the customer’s unique invoice_prefix if it is specified.",
                max_length=64,
            ),
        ),
        migrations.AlterField(
            model_name="invoice",
            name="statement_descriptor",
            field=models.CharField(
                blank=True,
                default="",
                help_text="An arbitrary string to be displayed on your customer's credit card statement. The statement description may not include <>\"' characters, and will appear on your customer's statement in capital letters. Non-ASCII characters are automatically stripped. While most banks display this information consistently, some may display it incorrectly or not at all.",
                max_length=22,
            ),
        ),
        migrations.AlterField(
            model_name="payout",
            name="failure_code",
            field=djstripe.fields.StripeEnumField(
                blank=True,
                default="",
                enum=djstripe.enums.PayoutFailureCode,
                help_text="Error code explaining reason for transfer failure if available. See https://stripe.com/docs/api/python#transfer_failures.",
                max_length=23,
            ),
        ),
        migrations.AlterField(
            model_name="payout",
            name="failure_message",
            field=models.TextField(
                blank=True,
                default="",
                help_text="Message to user further explaining reason for payout failure if available.",
            ),
        ),
        migrations.AlterField(
            model_name="payout",
            name="statement_descriptor",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Extra information about a payout to be displayed on the user's bank statement.",
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="plan",
            name="aggregate_usage",
            field=djstripe.fields.StripeEnumField(
                blank=True,
                default="",
                enum=djstripe.enums.PlanAggregateUsage,
                help_text="Specifies a usage aggregation strategy for plans of usage_type=metered. Allowed values are `sum` for summing up all usage during a period, `last_during_period` for picking the last usage record reported within a period, `last_ever` for picking the last usage record ever (across period bounds) or max which picks the usage record with the maximum reported usage during a period. Defaults to `sum`.",
                max_length=18,
            ),
        ),
        migrations.AlterField(
            model_name="plan",
            name="billing_scheme",
            field=djstripe.fields.StripeEnumField(
                blank=True,
                default="",
                enum=djstripe.enums.PlanBillingScheme,
                help_text="Describes how to compute the price per period. Either `per_unit` or `tiered`. `per_unit` indicates that the fixed amount (specified in amount) will be charged per unit in quantity (for plans with `usage_type=licensed`), or per unit of total usage (for plans with `usage_type=metered`). `tiered` indicates that the unit pricing will be computed using a tiering strategy as defined using the tiers and tiers_mode attributes.",
                max_length=8,
            ),
        ),
        migrations.AlterField(
            model_name="plan",
            name="nickname",
            field=models.TextField(
                blank=True,
                default="",
                help_text="A brief description of the plan, hidden from customers.",
                max_length=5000,
            ),
        ),
        migrations.AlterField(
            model_name="product",
            name="caption",
            field=models.TextField(
                blank=True,
                default="",
                help_text="A short one-line description of the product, meant to be displayableto the customer. Only applicable to products of `type=good`.",
                max_length=5000,
            ),
        ),
        migrations.AlterField(
            model_name="product",
            name="statement_descriptor",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Extra information about a product which will appear on your customer's credit card statement. In the case that multiple products are billed at once, the first statement descriptor will be used. Only available on products of type=`service`.",
                max_length=22,
            ),
        ),
        migrations.AlterField(
            model_name="product",
            name="unit_label",
            field=models.CharField(blank=True, default="", max_length=12),
        ),
        migrations.AlterField(
            model_name="refund",
            name="failure_reason",
            field=djstripe.fields.StripeEnumField(
                blank=True,
                default="",
                enum=djstripe.enums.RefundFailureReason,
                help_text="If the refund failed, the reason for refund failure if known.",
                max_length=24,
            ),
        ),
        migrations.AlterField(
            model_name="refund",
            name="reason",
            field=djstripe.fields.StripeEnumField(
                blank=True,
                default="",
                enum=djstripe.enums.RefundReason,
                help_text="Reason for the refund.",
                max_length=21,
            ),
        ),
        migrations.AlterField(
            model_name="refund",
            name="receipt_number",
            field=models.CharField(
                blank=True,
                default="",
                help_text="The transaction number that appears on email receipts sent for this charge.",
                max_length=9,
            ),
        ),
        migrations.AlterField(
            model_name="source",
            name="currency",
            field=djstripe.fields.StripeCurrencyCodeField(
                blank=True,
                default="",
                help_text="Three-letter ISO currency code",
                max_length=3,
            ),
        ),
        migrations.AlterField(
            model_name="source",
            name="statement_descriptor",
            field=models.CharField(
                blank=True,
                default="",
                help_text="Extra information about a source. This will appear on your customer's statement every time you charge the source.",
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="transfer",
            name="transfer_group",
            field=models.CharField(
                blank=True,
                default="",
                help_text="A string that identifies this transaction as part of a group.",
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="product",
            name="name",
            field=models.TextField(
                help_text="The product's name, meant to be displayable to the customer. Applicable to both `service` and `good` types.",
                max_length=5000,
            ),
        ),
        migrations.AlterField(
            model_name="subscription",
            name="plan",
            field=models.ForeignKey(
                blank=True,
                help_text="The plan associated with this subscription. This value will be `null` for multi-plan subscriptions",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="subscriptions",
                to="djstripe.Plan",
            ),
        ),
        migrations.AlterField(
            model_name="subscription",
            name="quantity",
            field=models.IntegerField(
                blank=True,
                help_text="The quantity applied to this subscription. This value will be `null` for multi-plan subscriptions",
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="plan",
            name="amount",
            field=djstripe.fields.StripeDecimalCurrencyAmountField(
                blank=True,
                decimal_places=2,
                help_text="Amount (as decimal) to be charged on the interval specified.",
                max_digits=8,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="invoice",
            name="closed",
            field=models.NullBooleanField(
                default=False,
                help_text="Whether or not the invoice is still trying to collect payment. An invoice is closed if it's either paid or it has been marked closed. A closed invoice will no longer attempt to collect payment.",
            ),
        ),
        migrations.AlterField(
            model_name="invoice",
            name="forgiven",
            field=models.NullBooleanField(
                default=False,
                help_text="Whether or not the invoice has been forgiven. Forgiving an invoice instructs us to update the subscription status as if the invoice were successfully paid. Once an invoice has been forgiven, it cannot be unforgiven or reopened.",
            ),
        ),
        migrations.RenameModel(
            old_name="PaymentMethod", new_name="DjstripePaymentMethod"
        ),
    ]



from typing import List

from django.apps import apps
from django.core.management.base import BaseCommand, CommandError

from ... import models, settings


class Command(BaseCommand):
    """Sync models from stripe."""

    help = "Sync models from stripe."

    def add_arguments(self, parser):
        parser.add_argument(
            "args",
            metavar="ModelName",
            nargs="*",
            help="restricts sync to these model names (default is to sync all "
            "supported models)",
        )

    def handle(self, *args, **options):
        app_label = "djstripe"
        app_config = apps.get_app_config(app_label)
        model_list = []  # type: List[models.StripeModel]

        if args:
            for model_label in args:
                try:
                    model = app_config.get_model(model_label)
                except LookupError:
                    raise CommandError(
                        "Unknown model: {}.{}".format(app_label, model_label)
                    )

                model_list.append(model)
        else:
            model_list = app_config.get_models()

        for model in model_list:
            self.sync_model(model)

    def sync_model(self, model):
        model_name = model.__name__

        if not issubclass(model, models.StripeModel):
            print("Skipping {} (not a StripeModel)".format(model_name))
            return

        if model.stripe_class is None:
            print("Skipping {} (no stripe_class)".format(model_name))
            return

        if not hasattr(model.stripe_class, "list"):
            print("Skipping {} (no stripe_class.list)".format(model_name))
            return

        print("Syncing {}:".format(model_name))

        try:
            count = 0

            for list_kwargs in self.get_list_kwargs(model):
                if model is models.Account:
                    # special case, since own account isn't returned by Account.api_list
                    stripe_obj = models.Account.stripe_class.retrieve(
                        api_key=settings.STRIPE_SECRET_KEY
                    )
                    count += 1
                    djstripe_obj = model.sync_from_stripe_data(stripe_obj)
                    print(
                        "  id={id}, pk={pk} ({djstripe_obj})".format(
                            id=djstripe_obj.id,
                            pk=djstripe_obj.pk,
                            djstripe_obj=djstripe_obj,
                        )
                    )

                for stripe_obj in model.api_list(**list_kwargs):
                    count += 1
                    djstripe_obj = model.sync_from_stripe_data(stripe_obj)
                    print(
                        "  id={id}, pk={pk} ({djstripe_obj})".format(
                            id=djstripe_obj.id,
                            pk=djstripe_obj.pk,
                            djstripe_obj=djstripe_obj,
                        )
                    )

            if count == 0:
                print("  (no results)")
            else:
                print(
                    "  Synced {count} {model_name}".format(
                        count=count, model_name=model_name
                    )
                )

        except Exception as e:
            print(e)

    def get_list_kwargs(self, model):
        """
        Returns a sequence of kwargs dicts to pass to model.api_list

        This allows us to sync models that require parameters to api_list

        :param model:
        :return: Sequence[dict]
        """
        if model is models.PaymentMethod:
            # special case
            all_list_kwargs = (
                {"customer": stripe_customer.id, "type": "card"}
                for stripe_customer in models.Customer.api_list()
            )
        else:
            # one empty dict so we iterate once
            all_list_kwargs = [{}]

        return all_list_kwargs



"""
dj-stripe middleware

Refer to SubscriptionPaymentMiddleware docstring for more info.
"""
import fnmatch

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.shortcuts import redirect
from django.urls import resolve
from django.utils.deprecation import MiddlewareMixin

from .settings import SUBSCRIPTION_REDIRECT, subscriber_request_callback
from .utils import subscriber_has_active_subscription

DJSTRIPE_SUBSCRIPTION_REQUIRED_EXCEPTION_URLS = getattr(
    settings, "DJSTRIPE_SUBSCRIPTION_REQUIRED_EXCEPTION_URLS", ()
)


# So we don't have crazy long lines of code
EXEMPT = list(DJSTRIPE_SUBSCRIPTION_REQUIRED_EXCEPTION_URLS)
EXEMPT.append("[djstripe]")


class SubscriptionPaymentMiddleware(MiddlewareMixin):
    """
    Used to redirect users from subcription-locked request destinations.

    Rules:

        * "(app_name)" means everything from this app is exempt
        * "[namespace]" means everything with this name is exempt
        * "namespace:name" means this namespaced URL is exempt
        * "name" means this URL is exempt
        * The entire djstripe namespace is exempt
        * If settings.DEBUG is True, then django-debug-toolbar is exempt
        * A 'fn:' prefix means the rest of the URL is fnmatch'd.

    Example::

        DJSTRIPE_SUBSCRIPTION_REQUIRED_EXCEPTION_URLS = (
            "[blogs]",  # Anything in the blogs namespace
            "products:detail",  # A ProductDetail view you want shown to non-payers
            "home",  # Site homepage
            "fn:/accounts*",  # anything in the accounts/ URL path
        )
    """

    def process_request(self, request):
        """Check the subscriber's subscription status.

        Returns early if request does not outlined in this middleware's docstring.
        """
        if self.is_matching_rule(request):
            return

        return self.check_subscription(request)

    def is_matching_rule(self, request):
        """Check according to the rules defined in the class docstring."""
        # First, if in DEBUG mode and with django-debug-toolbar, we skip
        #   this entire process.
        if settings.DEBUG and request.path.startswith("/__debug__"):
            return True

        # Second we check against matches
        match = resolve(
            request.path, getattr(request, "urlconf", settings.ROOT_URLCONF)
        )
        if "({0})".format(match.app_name) in EXEMPT:
            return True

        if "[{0}]".format(match.namespace) in EXEMPT:
            return True

        if "{0}:{1}".format(match.namespace, match.url_name) in EXEMPT:
            return True

        if match.url_name in EXEMPT:
            return True

        # Third, we check wildcards:
        for exempt in [x for x in EXEMPT if x.startswith("fn:")]:
            exempt = exempt.replace("fn:", "")
            if fnmatch.fnmatch(request.path, exempt):
                return True

        return False

    def check_subscription(self, request):
        """Redirect to the subscribe page if the user lacks an active subscription."""
        subscriber = subscriber_request_callback(request)

        if not subscriber_has_active_subscription(subscriber):
            if not SUBSCRIPTION_REDIRECT:
                raise ImproperlyConfigured("DJSTRIPE_SUBSCRIPTION_REDIRECT is not set.")
            return redirect(SUBSCRIPTION_REDIRECT)



"""
Webhook event handlers for the various models

Stripe docs for Events: https://stripe.com/docs/api/events
Stripe docs for Webhooks: https://stripe.com/docs/webhooks

TODO: Implement webhook event handlers for all the models that need to
      respond to webhook events.

NOTE:
    Event data is not guaranteed to be in the correct API version format.
    See #116. When writing a webhook handler, make sure to first
    re-retrieve the object you wish to process.

"""
import logging

from . import models, webhooks
from .enums import SourceType
from .utils import convert_tstamp

logger = logging.getLogger(__name__)


@webhooks.handler("customer")
def customer_webhook_handler(event):
    """Handle updates to customer objects.

    First determines the crud_type and then handles the event if a customer
    exists locally.
    As customers are tied to local users, djstripe will not create customers that
    do not already exist locally.

    Docs and an example customer webhook response:
    https://stripe.com/docs/api#customer_object
    """
    if event.customer:
        # As customers are tied to local users, djstripe will not create
        # customers that do not already exist locally.
        _handle_crud_like_event(
            target_cls=models.Customer, event=event, crud_exact=True, crud_valid=True
        )


@webhooks.handler("customer.discount")
def customer_discount_webhook_handler(event):
    """Handle updates to customer discount objects.

    Docs: https://stripe.com/docs/api#discounts

    Because there is no concept of a "Discount" model in dj-stripe (due to the
    lack of a stripe id on them), this is a little different to the other
    handlers.
    """

    crud_type = CrudType.determine(event=event)
    discount_data = event.data.get("object", {})
    coupon_data = discount_data.get("coupon", {})
    customer = event.customer

    if crud_type.created or crud_type.updated:
        coupon, _ = _handle_crud_like_event(
            target_cls=models.Coupon,
            event=event,
            data=coupon_data,
            id=coupon_data.get("id"),
        )
        coupon_start = discount_data.get("start")
        coupon_end = discount_data.get("end")
    else:
        coupon = None
        coupon_start = None
        coupon_end = None

    customer.coupon = coupon
    customer.coupon_start = convert_tstamp(coupon_start)
    customer.coupon_end = convert_tstamp(coupon_end)
    customer.save()


@webhooks.handler("customer.source")
def customer_source_webhook_handler(event):
    """Handle updates to customer payment-source objects.

    Docs: https://stripe.com/docs/api#customer_object-sources.
    """
    customer_data = event.data.get("object", {})
    source_type = customer_data.get("object", {})

    # TODO: handle other types of sources
    #  (https://stripe.com/docs/api#customer_object-sources)
    if source_type == SourceType.card:
        if event.verb.endswith("deleted") and customer_data:
            # On customer.source.deleted, we do not delete the object,
            # we merely unlink it.
            # customer = Customer.objects.get(id=customer_data["id"])
            # NOTE: for now, customer.sources still points to Card
            # Also, https://github.com/dj-stripe/dj-stripe/issues/576
            models.Card.objects.filter(id=customer_data.get("id", "")).delete()
            models.DjstripePaymentMethod.objects.filter(
                id=customer_data.get("id", "")
            ).delete()
        else:
            _handle_crud_like_event(target_cls=models.Card, event=event)


@webhooks.handler("customer.subscription")
def customer_subscription_webhook_handler(event):
    """Handle updates to customer subscription objects.

    Docs an example subscription webhook response:
    https://stripe.com/docs/api#subscription_object
    """

    # customer.subscription.deleted doesn't actually delete the subscription
    # on the stripe side, it updates it to canceled status, so override
    # crud_type to update to match.
    crud_type = CrudType.determine(event=event)
    if crud_type.deleted:
        crud_type = CrudType(updated=True)
    _handle_crud_like_event(
        target_cls=models.Subscription, event=event, crud_type=crud_type
    )


@webhooks.handler("payment_method")
def payment_method_handler(event):
    """
    Handle updates to payment_method objects
    :param event:
    :return:

    Docs for:
    - payment_method: https://stripe.com/docs/api/payment_methods
    """
    id_ = event.data.get("object", {}).get("id", None)

    if (
        event.parts == ["payment_method", "detached"]
        and id_
        and id_.startswith("card_")
    ):
        # Special case to handle a quirk in stripe's wrapping of legacy "card" objects
        # with payment_methods - card objects are deleted on detach, so treat this as
        # a delete event
        _handle_crud_like_event(
            target_cls=models.PaymentMethod,
            event=event,
            crud_type=CrudType(deleted=True),
        )
    else:
        _handle_crud_like_event(target_cls=models.PaymentMethod, event=event)


@webhooks.handler(
    "charge",
    "coupon",
    "invoice",
    "invoiceitem",
    "payment_intent",
    "plan",
    "product",
    "setup_intent",
    "source",
    "tax_rate",
    "transfer",
)
def other_object_webhook_handler(event):
    """
    Handle updates to charge, coupon, invoice, invoiceitem, payment_intent,
    plan, product, setup_intent, source, tax_rate and transfer objects.

    Docs for:
    - charge: https://stripe.com/docs/api/charges
    - coupon: https://stripe.com/docs/api/coupons
    - invoice: https://stripe.com/docs/api/invoices
    - invoiceitem: https://stripe.com/docs/api/invoiceitems
    - payment_intent: https://stripe.com/docs/api/payment_intents
    - plan: https://stripe.com/docs/api/plans
    - product: https://stripe.com/docs/api/products
    - setup_intent: https://stripe.com/docs/api/setup_intents
    - source: https://stripe.com/docs/api/sources
    - tax_rate: https://stripe.com/docs/api/tax_rates/
    - transfer: https://stripe.com/docs/api/transfers
    """

    if event.parts[:2] == ["charge", "dispute"]:
        # Do not attempt to handle charge.dispute.* events.
        # We do not have a Dispute model yet.
        target_cls = models.Dispute
    else:
        target_cls = {
            "charge": models.Charge,
            "coupon": models.Coupon,
            "invoice": models.Invoice,
            "invoiceitem": models.InvoiceItem,
            "payment_intent": models.PaymentIntent,
            "plan": models.Plan,
            "product": models.Product,
            "transfer": models.Transfer,
            "setup_intent": models.SetupIntent,
            "source": models.Source,
            "tax_rate": models.TaxRate,
        }.get(event.category)

    _handle_crud_like_event(target_cls=target_cls, event=event)


#
# Helpers
#


class CrudType(object):
    """Helper object to determine CRUD-like event state."""

    created = False
    updated = False
    deleted = False

    def __init__(self, **kwargs):
        """Set attributes."""
        for k, v in kwargs.items():
            setattr(self, k, v)

    @property
    def valid(self):
        """Return True if this is a CRUD-like event."""
        return self.created or self.updated or self.deleted

    @classmethod
    def determine(cls, event, verb=None, exact=False):
        """
        Determine if the event verb is a crud_type (without the 'R') event.

        :param event:
        :type event: models.Event
        :param verb: The event verb to examine.
        :type verb: str
        :param exact: If True, match crud_type to event verb string exactly.
        :type exact: bool
        :returns: The CrudType state object.
        :rtype: CrudType
        """
        verb = verb or event.verb

        def check(crud_type_event):
            if exact:
                return verb == crud_type_event
            else:
                return verb.endswith(crud_type_event)

        created = updated = deleted = False

        if check("updated"):
            updated = True
        elif check("created"):
            created = True
        elif check("deleted"):
            deleted = True

        return cls(created=created, updated=updated, deleted=deleted)


def _handle_crud_like_event(
    target_cls,
    event,
    data=None,
    verb=None,
    id=None,
    customer=None,
    crud_type=None,
    crud_exact=False,
    crud_valid=False,
):
    """
    Helper to process crud_type-like events for objects.

    Non-deletes (creates, updates and "anything else" events) are treated as
    update_or_create events - The object will be retrieved locally, then it is
    synchronised with the Stripe API for parity.

    Deletes only occur for delete events and cause the object to be deleted
    from the local database, if it existed.  If it doesn't exist then it is
    ignored (but the event processing still succeeds).

    :param target_cls: The djstripe model being handled.
    :type target_cls: Type[models.StripeModel]
    :param event: The event object
    :type event: models.Event
    :param data: The event object data (defaults to ``event.data``).
    :param verb: The event verb (defaults to ``event.verb``).
    :type verb: str
    :param id: The object Stripe ID (defaults to ``object.id``).
    :type id: str
    :param customer: The customer object (defaults to ``event.customer``).
    :param crud_type: The CrudType object (determined by default).
    :param crud_exact: If True, match verb against CRUD type exactly.
    :param crud_valid: If True, CRUD type must match valid type.
    :returns: The object (if any) and the event CrudType.
    :rtype: Tuple[models.StripeModel, CrudType]
    """
    data = data or event.data
    id = id or data.get("object", {}).get("id", None)

    if not id:
        # We require an object when applying CRUD-like events, so if there's
        # no ID the event is ignored/dropped. This happens in events such as
        # invoice.upcoming, which refer to a future (non-existant) invoice.
        logger.debug(
            "Ignoring %r Stripe event without object ID: %r", event.type, event
        )
        return

    verb = verb or event.verb
    customer = customer or event.customer
    crud_type = crud_type or CrudType.determine(
        event=event, verb=verb, exact=crud_exact
    )
    obj = None

    if crud_valid and not crud_type.valid:
        logger.debug(
            "Ignoring %r Stripe event without valid CRUD type: %r", event.type, event
        )
        return

    if crud_type.deleted:
        qs = target_cls.objects.filter(id=id)
        if target_cls is models.Customer and qs.exists():
            qs.get().purge()
        else:
            obj = target_cls.objects.filter(id=id).delete()
    else:
        # Any other event type (creates, updates, etc.) - This can apply to
        # verbs that aren't strictly CRUD but Stripe do intend an update.  Such
        # as invoice.payment_failed.
        kwargs = {"id": id}
        if hasattr(target_cls, "customer"):
            kwargs["customer"] = customer
        data = target_cls(**kwargs).api_retrieve()
        obj = target_cls.sync_from_stripe_data(data)

    return obj, crud_type



"""
.. module:: dj-stripe.contrib.rest_framework.urls.

    :synopsis: URL routes for the dj-stripe REST API.

.. moduleauthor:: Philippe Luickx (@philippeluickx)

Wire this into the root URLConf this way::

    path(
        'api/v1/stripe/',
        include('djstripe.contrib.rest_framework.urls', namespace="rest_djstripe")
    ),
    # url can be changed
    # Call to 'djstripe.contrib.rest_framework.urls' and 'namespace' must stay as is

"""

from django.urls import path

from . import views

app_name = "djstripe_rest_framework"

urlpatterns = [
    # REST api
    path("subscription/", views.SubscriptionRestView.as_view(), name="subscription")
]



import stripe
from django.db import models

from .. import enums
from .. import settings as djstripe_settings
from ..fields import (
    JSONField,
    StripeCurrencyCodeField,
    StripeDecimalCurrencyAmountField,
    StripeEnumField,
    StripeIdField,
    StripeQuantumCurrencyAmountField,
)
from ..managers import TransferManager
from .base import StripeModel


class Account(StripeModel):
    """
    Stripe documentation: https://stripe.com/docs/api#account
    """

    stripe_class = stripe.Account
    # Special handling of the icon and logo fields, they moved to settings.branding
    # in Stripe 2019-02-19 but we want them as ForeignKeys
    branding_icon = models.ForeignKey(
        "FileUpload",
        on_delete=models.SET_NULL,
        null=True,
        related_name="icon_account",
        help_text="An icon for the account. Must be square and at least 128px x 128px.",
    )
    branding_logo = models.ForeignKey(
        "FileUpload",
        on_delete=models.SET_NULL,
        null=True,
        related_name="logo_account",
        help_text="A logo for the account that will be used in Checkout instead of "
        "the icon and without the account’s name next to it if provided. "
        "Must be at least 128px x 128px.",
    )
    business_profile = JSONField(
        null=True, blank=True, help_text="Optional information related to the business."
    )
    business_type = StripeEnumField(
        enum=enums.BusinessType, default="", blank=True, help_text="The business type."
    )
    charges_enabled = models.BooleanField(
        help_text="Whether the account can create live charges"
    )
    country = models.CharField(max_length=2, help_text="The country of the account")
    company = JSONField(
        null=True,
        blank=True,
        help_text=(
            "Information about the company or business. "
            "This field is null unless business_type is set to company."
        ),
    )
    default_currency = StripeCurrencyCodeField(
        help_text="The currency this account has chosen to use as the default"
    )
    details_submitted = models.BooleanField(
        help_text=(
            "Whether account details have been submitted. "
            "Standard accounts cannot receive payouts before this is true."
        )
    )
    email = models.CharField(
        max_length=255, help_text="The primary user’s email address."
    )
    # TODO external_accounts = ...
    individual = JSONField(
        null=True,
        blank=True,
        help_text=(
            "Information about the person represented by the account. "
            "This field is null unless business_type is set to individual."
        ),
    )
    payouts_enabled = models.BooleanField(
        help_text="Whether Stripe can send payouts to this account"
    )
    product_description = models.CharField(
        max_length=255,
        default="",
        blank=True,
        help_text="Internal-only description of the product sold or service provided "
        "by the business. It’s used by Stripe for risk and underwriting purposes.",
    )
    requirements = JSONField(
        null=True,
        blank=True,
        help_text="Information about the requirements for the account, "
        "including what information needs to be collected, and by when.",
    )
    settings = JSONField(
        null=True,
        blank=True,
        help_text=(
            "Account options for customizing how the account functions within Stripe."
        ),
    )
    type = StripeEnumField(enum=enums.AccountType, help_text="The Stripe account type.")
    tos_acceptance = JSONField(
        null=True,
        blank=True,
        help_text="Details on the acceptance of the Stripe Services Agreement",
    )

    @property
    def business_url(self):
        """
        The business’s publicly available website.
        :rtype: Optional[str]
        """
        return (self.business_profile or {}).get("url")

    @classmethod
    def get_connected_account_from_token(cls, access_token):
        account_data = cls.stripe_class.retrieve(api_key=access_token)

        return cls._get_or_create_from_stripe_object(account_data)[0]

    @classmethod
    def get_default_account(cls):
        account_data = cls.stripe_class.retrieve(
            api_key=djstripe_settings.STRIPE_SECRET_KEY
        )

        return cls._get_or_create_from_stripe_object(account_data)[0]

    def __str__(self):
        return (
            self.settings.get("dashboard", {}).get("display_name")
            or self.business_profile.get("name")
            or super().__str__()
        )

    @classmethod  # noqa: C901
    def _manipulate_stripe_object_hook(cls, data):
        data = super()._manipulate_stripe_object_hook(data)

        def empty_string_to_none(v):
            """
            stripe.StripeObject.__setitem__ doesn't allow = ""
            """
            if v == "":
                return None
            else:
                return v

        # icon (formerly called business_logo)
        # logo (formerly called business_logo_large)
        # moved to settings.branding in Stripe 2019-02-19
        # but we'll keep them to provide the ForeignKey
        for old, new in [("branding_icon", "icon"), ("branding_logo", "logo")]:
            try:
                data[old] = data["settings"]["branding"][new]
            except KeyError:
                pass

        return data

    @classmethod
    def _create_from_stripe_object(
        cls,
        data,
        current_ids=None,
        pending_relations=None,
        save=True,
        stripe_account=None,
    ):
        """
        Set the stripe_account to the id of the Account instance being created.

        This ensures that the foreign-key relations that may exist in stripe are
        fetched using the appropriate connected account ID.
        """
        return super()._create_from_stripe_object(
            data=data,
            current_ids=current_ids,
            pending_relations=pending_relations,
            save=save,
            stripe_account=data["id"] if not stripe_account else stripe_account,
        )


class ApplicationFee(StripeModel):
    """
    When you collect a transaction fee on top of a charge made for your
    user (using Connect), an ApplicationFee is created in your account.

    Stripe documentation: https://stripe.com/docs/api#application_fees
    """

    stripe_class = stripe.ApplicationFee

    amount = StripeQuantumCurrencyAmountField(help_text="Amount earned, in cents.")
    amount_refunded = StripeQuantumCurrencyAmountField(
        help_text="Amount in cents refunded (can be less than the amount attribute "
        "on the fee if a partial refund was issued)"
    )
    # TODO application = ...
    balance_transaction = models.ForeignKey(
        "BalanceTransaction",
        on_delete=models.CASCADE,
        help_text="Balance transaction that describes the impact on your account"
        " balance.",
    )
    charge = models.ForeignKey(
        "Charge",
        on_delete=models.CASCADE,
        help_text="The charge that the application fee was taken from.",
    )
    currency = StripeCurrencyCodeField()
    # TODO originating_transaction = ... (refs. both Charge and Transfer)
    refunded = models.BooleanField(
        help_text=(
            "Whether the fee has been fully refunded. If the fee is only "
            "partially refunded, this attribute will still be false."
        )
    )


class ApplicationFeeRefund(StripeModel):
    """
    ApplicationFeeRefund objects allow you to refund an ApplicationFee that
    has previously been created but not yet refunded.
    Funds will be refunded to the Stripe account from which the fee was
    originally collected.

    Stripe documentation: https://stripe.com/docs/api#fee_refunds
    """

    description = None

    amount = StripeQuantumCurrencyAmountField(help_text="Amount refunded, in cents.")
    balance_transaction = models.ForeignKey(
        "BalanceTransaction",
        on_delete=models.CASCADE,
        help_text="Balance transaction that describes the impact on your account "
        "balance.",
    )
    currency = StripeCurrencyCodeField()
    fee = models.ForeignKey(
        "ApplicationFee",
        on_delete=models.CASCADE,
        related_name="refunds",
        help_text="The application fee that was refunded",
    )


class CountrySpec(StripeModel):
    """
    Stripe documentation: https://stripe.com/docs/api#country_specs
    """

    stripe_class = stripe.CountrySpec

    id = models.CharField(max_length=2, primary_key=True, serialize=True)

    default_currency = StripeCurrencyCodeField(
        help_text=(
            "The default currency for this country. "
            "This applies to both payment methods and bank accounts."
        )
    )
    supported_bank_account_currencies = JSONField(
        help_text="Currencies that can be accepted in the specific country"
        " (for transfers)."
    )
    supported_payment_currencies = JSONField(
        help_text="Currencies that can be accepted in the specified country"
        " (for payments)."
    )
    supported_payment_methods = JSONField(
        help_text="Payment methods available in the specified country."
    )
    supported_transfer_countries = JSONField(
        help_text="Countries that can accept transfers from the specified country."
    )
    verification_fields = JSONField(
        help_text="Lists the types of verification data needed to keep an account open."
    )

    # Get rid of core common fields
    djstripe_id = None
    created = None
    description = None
    livemode = True
    metadata = None

    class Meta:
        pass


class Transfer(StripeModel):
    """
    When Stripe sends you money or you initiate a transfer to a bank account,
    debit card, or connected Stripe account, a transfer object will be created.

    Stripe documentation: https://stripe.com/docs/api/python#transfers
    """

    stripe_class = stripe.Transfer
    expand_fields = ["balance_transaction"]
    stripe_dashboard_item_name = "transfers"

    objects = TransferManager()

    amount = StripeDecimalCurrencyAmountField(help_text="The amount transferred")
    amount_reversed = StripeDecimalCurrencyAmountField(
        null=True,
        blank=True,
        help_text="The amount (as decimal) reversed (can be less than the amount "
        "attribute on the transfer if a partial reversal was issued).",
    )
    balance_transaction = models.ForeignKey(
        "BalanceTransaction",
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        help_text="Balance transaction that describes the impact on your account"
        " balance.",
    )
    currency = StripeCurrencyCodeField()
    # TODO: Link destination to Card, Account, or Bank Account Models
    destination = StripeIdField(
        help_text="ID of the bank account, card, or Stripe account the transfer was "
        "sent to."
    )
    destination_payment = StripeIdField(
        null=True,
        blank=True,
        help_text="If the destination is a Stripe account, this will be the ID of the "
        "payment that the destination account received for the transfer.",
    )
    reversed = models.BooleanField(
        default=False,
        help_text="Whether or not the transfer has been fully reversed. "
        "If the transfer is only partially reversed, this attribute will still "
        "be false.",
    )
    source_transaction = StripeIdField(
        null=True,
        help_text="ID of the charge (or other transaction) that was used to fund "
        "the transfer. If null, the transfer was funded from the available balance.",
    )
    source_type = StripeEnumField(
        enum=enums.LegacySourceType,
        help_text="The source balance from which this transfer came.",
    )
    transfer_group = models.CharField(
        max_length=255,
        default="",
        blank=True,
        help_text="A string that identifies this transaction as part of a group.",
    )

    @property
    def fee(self):
        if self.balance_transaction:
            return self.balance_transaction.fee

    def str_parts(self):
        return ["amount={amount}".format(amount=self.amount)] + super().str_parts()


class TransferReversal(StripeModel):
    """
    Stripe documentation: https://stripe.com/docs/api#transfer_reversals
    """

    stripe_class = stripe.Transfer

    amount = StripeQuantumCurrencyAmountField(help_text="Amount, in cents.")
    balance_transaction = models.ForeignKey(
        "BalanceTransaction",
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        related_name="transfer_reversals",
        help_text="Balance transaction that describes the impact on your account "
        "balance.",
    )
    currency = StripeCurrencyCodeField()
    transfer = models.ForeignKey(
        "Transfer",
        on_delete=models.CASCADE,
        help_text="The transfer that was reversed.",
        related_name="reversals",
    )



"""
dj-stripe Decorators.
"""
from functools import wraps

from django.core.exceptions import ImproperlyConfigured
from django.shortcuts import redirect

from .settings import SUBSCRIPTION_REDIRECT, subscriber_request_callback
from .utils import subscriber_has_active_subscription


def subscriber_passes_pay_test(test_func, plan=None, pay_page=SUBSCRIPTION_REDIRECT):
    """
    Decorator for views that checks the subscriber passes the given test for a
    "Paid Feature".

    Redirects to `pay_page` if necessary. The test should be a callable
    that takes the subscriber object and returns True if the subscriber passes.
    """

    def decorator(view_func):
        @wraps(view_func)
        def _wrapped_view(request, *args, **kwargs):
            if test_func(subscriber_request_callback(request), plan):
                return view_func(request, *args, **kwargs)

            if not pay_page:
                raise ImproperlyConfigured("DJSTRIPE_SUBSCRIPTION_REDIRECT is not set.")

            return redirect(pay_page)

        return _wrapped_view

    return decorator


def subscription_payment_required(
    function=None, plan=None, pay_page=SUBSCRIPTION_REDIRECT
):
    """
    Decorator for views that require subscription payment.

    Redirects to `pay_page` if necessary.
    """
    actual_decorator = subscriber_passes_pay_test(
        subscriber_has_active_subscription, plan=plan, pay_page=pay_page
    )
    if function:
        return actual_decorator(function)
    return actual_decorator



"""
dj-stripe System Checks
"""
from django.conf import settings
from django.core import checks
from django.utils.dateparse import date_re


@checks.register("djstripe")
def check_stripe_api_key(app_configs=None, **kwargs):
    """Check the user has configured API live/test keys correctly."""
    from . import settings as djstripe_settings

    messages = []

    if not djstripe_settings.STRIPE_SECRET_KEY:
        msg = "Could not find a Stripe API key."
        hint = "Add STRIPE_TEST_SECRET_KEY and STRIPE_LIVE_SECRET_KEY to your settings."
        messages.append(checks.Critical(msg, hint=hint, id="djstripe.C001"))
    elif djstripe_settings.STRIPE_LIVE_MODE:
        if not djstripe_settings.LIVE_API_KEY.startswith(("sk_live_", "rk_live_")):
            msg = "Bad Stripe live API key."
            hint = 'STRIPE_LIVE_SECRET_KEY should start with "sk_live_"'
            messages.append(checks.Critical(msg, hint=hint, id="djstripe.C002"))
    else:
        if not djstripe_settings.TEST_API_KEY.startswith(("sk_test_", "rk_test_")):
            msg = "Bad Stripe test API key."
            hint = 'STRIPE_TEST_SECRET_KEY should start with "sk_test_"'
            messages.append(checks.Critical(msg, hint=hint, id="djstripe.C003"))

    return messages


def validate_stripe_api_version(version):
    """
    Check the API version is formatted correctly for Stripe.

    The expected format is an iso8601 date: `YYYY-MM-DD`

    :param version: The version to set for the Stripe API.
    :type version: ``str``
    :returns bool: Whether the version is formatted correctly.
    """
    return date_re.match(version)


@checks.register("djstripe")
def check_stripe_api_version(app_configs=None, **kwargs):
    """Check the user has configured API version correctly."""
    from . import settings as djstripe_settings

    messages = []
    default_version = djstripe_settings.DEFAULT_STRIPE_API_VERSION
    version = djstripe_settings.get_stripe_api_version()

    if not validate_stripe_api_version(version):
        msg = "Invalid Stripe API version: {}".format(version)
        hint = "STRIPE_API_VERSION should be formatted as: YYYY-MM-DD"
        messages.append(checks.Critical(msg, hint=hint, id="djstripe.C004"))

    if version != default_version:
        msg = (
            "The Stripe API version has a non-default value of '{}'. "
            "Non-default versions are not explicitly supported, and may "
            "cause compatibility issues.".format(version)
        )
        hint = "Use the dj-stripe default for Stripe API version: {}".format(
            default_version
        )
        messages.append(checks.Warning(msg, hint=hint, id="djstripe.W001"))

    return messages


@checks.register("djstripe")
def check_native_jsonfield_postgres_engine(app_configs=None, **kwargs):
    """
    Check that the DJSTRIPE_USE_NATIVE_JSONFIELD isn't set unless Postgres is in use.
    """
    from . import settings as djstripe_settings

    messages = []
    error_msg = (
        "DJSTRIPE_USE_NATIVE_JSONFIELD is not compatible with engine {engine} "
        "for database {name}"
    )

    if djstripe_settings.USE_NATIVE_JSONFIELD:
        for db_name, db_config in settings.DATABASES.items():
            # Hi there.
            # You may be reading this because you are using Postgres, but
            # dj-stripe is not detecting that correctly. For example, maybe you
            # are using multiple databases with different engines, or you have
            # your own backend. As long as you are certain you can support jsonb,
            # you can use the SILENCED_SYSTEM_CHECKS setting to ignore this check.
            engine = db_config.get("ENGINE", "")
            if "postgresql" not in engine and "postgis" not in engine:
                messages.append(
                    checks.Critical(
                        error_msg.format(name=repr(db_name), engine=repr(engine)),
                        hint="Switch to Postgres, or unset "
                        "DJSTRIPE_USE_NATIVE_JSONFIELD",
                        id="djstripe.C005",
                    )
                )

    return messages


@checks.register("djstripe")
def check_stripe_api_host(app_configs=None, **kwargs):
    """
    Check that STRIPE_API_HOST is not being used in production.
    """
    from django.conf import settings

    messages = []

    if not settings.DEBUG and hasattr(settings, "STRIPE_API_HOST"):
        messages.append(
            checks.Warning(
                "STRIPE_API_HOST should not be set in production! "
                "This is most likely unintended.",
                hint="Remove STRIPE_API_HOST from your Django settings.",
                id="djstripe.W002",
            )
        )

    return messages


@checks.register("djstripe")
def check_webhook_secret(app_configs=None, **kwargs):
    """
    Check that DJSTRIPE_WEBHOOK_SECRET looks correct
    """
    from . import settings as djstripe_settings

    messages = []

    secret = djstripe_settings.WEBHOOK_SECRET
    if secret and not secret.startswith("whsec_"):
        messages.append(
            checks.Warning(
                "DJSTRIPE_WEBHOOK_SECRET does not look valid",
                hint="It should start with whsec_...",
                id="djstripe.W003",
            )
        )

    return messages


@checks.register("djstripe")
def check_webhook_validation(app_configs=None, **kwargs):
    """
    Check that DJSTRIPE_WEBHOOK_VALIDATION is valid
    """
    from . import settings as djstripe_settings

    messages = []

    validation_options = ("verify_signature", "retrieve_event")

    if djstripe_settings.WEBHOOK_VALIDATION is None:
        messages.append(
            checks.Warning(
                "Webhook validation is disabled, this is a security risk if the "
                "webhook view is enabled",
                hint="Set DJSTRIPE_WEBHOOK_VALIDATION to one of {}".format(
                    ", ".join(validation_options)
                ),
                id="djstripe.W004",
            )
        )
    elif djstripe_settings.WEBHOOK_VALIDATION == "verify_signature":
        if not djstripe_settings.WEBHOOK_SECRET:
            messages.append(
                checks.Critical(
                    "DJSTRIPE_WEBHOOK_VALIDATION='verify_signature' "
                    "but DJSTRIPE_WEBHOOK_SECRET is not set",
                    hint="Set DJSTRIPE_WEBHOOK_SECRET or set "
                    "DJSTRIPE_WEBHOOK_VALIDATION='retrieve_event'",
                    id="djstripe.C006",
                )
            )
    elif djstripe_settings.WEBHOOK_VALIDATION not in validation_options:
        messages.append(
            checks.Critical(
                "DJSTRIPE_WEBHOOK_VALIDATION is invalid",
                hint="Set DJSTRIPE_WEBHOOK_VALIDATION to one of {} or None".format(
                    ", ".join(validation_options)
                ),
                id="djstripe.C007",
            )
        )

    return messages


@checks.register("djstripe")
def check_subscriber_key_length(app_configs=None, **kwargs):
    """
    Check that DJSTRIPE_SUBSCRIBER_CUSTOMER_KEY fits in metadata.

    Docs: https://stripe.com/docs/api#metadata
    """
    from . import settings as djstripe_settings

    messages = []

    key = djstripe_settings.SUBSCRIBER_CUSTOMER_KEY
    key_size = len(str(key))
    if key and key_size > 40:
        messages.append(
            checks.Error(
                "DJSTRIPE_SUBSCRIBER_CUSTOMER_KEY must be no more than "
                "40 characters long",
                hint="Current value: %r (%i characters)" % (key, key_size),
                id="djstripe.E001",
            )
        )

    return messages



"""
Django Administration interface definitions
"""
from django.contrib import admin

from . import models


def get_forward_relation_fields_for_model(model):
    """Return an iterable of the field names that are forward relations,
    I.E ManyToManyField, OneToOneField, and ForeignKey.

    Useful for perhaps ensuring the admin is always using raw ID fields for
    newly added forward relation fields.
    """
    return [
        field.name
        for field in model._meta.get_fields()
        # Get only relation fields
        if field.is_relation
        # Exclude auto relation fields, like reverse one to one.
        and not field.auto_created
        # We only want forward relations.
        and any((field.many_to_many, field.one_to_one, field.many_to_one))
    ]


class BaseHasSourceListFilter(admin.SimpleListFilter):
    title = "source presence"
    parameter_name = "has_source"

    def lookups(self, request, model_admin):
        """
        Return a list of tuples.

        The first element in each tuple is the coded value for the option that will
        appear in the URL query. The second element is the
        human-readable name for the option that will appear
        in the right sidebar.
        source:
        https://docs.djangoproject.com/en/1.10/ref/contrib/admin/#django.contrib.admin.ModelAdmin.list_filter
        """
        return (("yes", "Has a source"), ("no", "Has no source"))

    def queryset(self, request, queryset):
        """
        Return the filtered queryset based on the value provided in the query string.

        source:
        https://docs.djangoproject.com/en/1.10/ref/contrib/admin/#django.contrib.admin.ModelAdmin.list_filter
        """
        filter_args = {self._filter_arg_key: None}

        if self.value() == "yes":
            return queryset.exclude(**filter_args)
        if self.value() == "no":
            return queryset.filter(**filter_args)


class CustomerHasSourceListFilter(BaseHasSourceListFilter):
    _filter_arg_key = "default_source"


class InvoiceCustomerHasSourceListFilter(BaseHasSourceListFilter):
    _filter_arg_key = "customer__default_source"


class CustomerSubscriptionStatusListFilter(admin.SimpleListFilter):
    """A SimpleListFilter used with Customer admin."""

    title = "subscription status"
    parameter_name = "sub_status"

    def lookups(self, request, model_admin):
        """
        Return a list of tuples.

        The first element in each tuple is the coded value for the option that will
        appear in the URL query. The second element is the
        human-readable name for the option that will appear
        in the right sidebar.
        source:
        https://docs.djangoproject.com/en/1.10/ref/contrib/admin/#django.contrib.admin.ModelAdmin.list_filter
        """
        statuses = [
            [x, x.replace("_", " ").title()]
            for x in models.Subscription.objects.values_list(
                "status", flat=True
            ).distinct()
        ]
        statuses.append(["none", "No Subscription"])
        return statuses

    def queryset(self, request, queryset):
        """
        Return the filtered queryset based on the value provided in the query string.

        source:
        https://docs.djangoproject.com/en/1.10/ref/contrib/admin/#django.contrib.admin.ModelAdmin.list_filter
        """
        if self.value() is None:
            return queryset.all()
        else:
            return queryset.filter(subscriptions__status=self.value()).distinct()


@admin.register(models.IdempotencyKey)
class IdempotencyKeyAdmin(admin.ModelAdmin):
    list_display = ("uuid", "action", "created", "is_expired", "livemode")
    list_filter = ("livemode",)
    search_fields = ("uuid", "action")

    def has_add_permission(self, request):
        return False


@admin.register(models.WebhookEventTrigger)
class WebhookEventTriggerAdmin(admin.ModelAdmin):
    list_display = (
        "created",
        "event",
        "remote_ip",
        "processed",
        "valid",
        "exception",
        "djstripe_version",
    )
    list_filter = ("created", "valid", "processed")
    raw_id_fields = get_forward_relation_fields_for_model(models.WebhookEventTrigger)

    def reprocess(self, request, queryset):
        for trigger in queryset:
            if not trigger.valid:
                self.message_user(request, "Skipped invalid trigger {}".format(trigger))
                continue

            trigger.process()

    def has_add_permission(self, request):
        return False


class StripeModelAdmin(admin.ModelAdmin):
    """Base class for all StripeModel-based model admins"""

    change_form_template = "djstripe/admin/change_form.html"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.raw_id_fields = get_forward_relation_fields_for_model(self.model)

    def get_list_display(self, request):
        return ("id",) + self.list_display + ("created", "livemode")

    def get_list_filter(self, request):
        return self.list_filter + ("created", "livemode")

    def get_readonly_fields(self, request, obj=None):
        return self.readonly_fields + ("id", "created")

    def get_search_fields(self, request):
        return self.search_fields + ("id",)

    def get_fieldsets(self, request, obj=None):
        common_fields = ("livemode", "id", "created")
        # Have to remove the fields from the common set,
        # otherwise they'll show up twice.
        fields = [f for f in self.get_fields(request, obj) if f not in common_fields]
        return (
            (None, {"fields": common_fields}),
            (self.model.__name__, {"fields": fields}),
        )


class SubscriptionInline(admin.StackedInline):
    """A TabularInline for use models.Subscription."""

    model = models.Subscription
    extra = 0
    readonly_fields = ("id", "created")
    raw_id_fields = get_forward_relation_fields_for_model(model)
    show_change_link = True


class SubscriptionItemInline(admin.StackedInline):
    """A TabularInline for use models.Subscription."""

    model = models.SubscriptionItem
    extra = 0
    readonly_fields = ("id", "created")
    raw_id_fields = get_forward_relation_fields_for_model(model)
    show_change_link = True


class InvoiceItemInline(admin.StackedInline):
    """A TabularInline for use InvoiceItem."""

    model = models.InvoiceItem
    extra = 0
    readonly_fields = ("id", "created")
    raw_id_fields = get_forward_relation_fields_for_model(model)
    show_change_link = True


@admin.register(models.Account)
class AccountAdmin(StripeModelAdmin):
    list_display = ("business_url", "country", "default_currency")
    list_filter = ("details_submitted",)
    search_fields = ("settings", "business_profile")


@admin.register(models.Charge)
class ChargeAdmin(StripeModelAdmin):
    list_display = (
        "customer",
        "amount",
        "description",
        "paid",
        "disputed",
        "refunded",
        "fee",
    )
    search_fields = ("customer__id", "invoice__id")
    list_filter = ("status", "paid", "refunded", "captured")


@admin.register(models.Coupon)
class CouponAdmin(StripeModelAdmin):
    list_display = (
        "amount_off",
        "percent_off",
        "duration",
        "duration_in_months",
        "redeem_by",
        "max_redemptions",
        "times_redeemed",
    )
    list_filter = ("duration", "redeem_by")
    radio_fields = {"duration": admin.HORIZONTAL}


@admin.register(models.Customer)
class CustomerAdmin(StripeModelAdmin):
    list_display = (
        "subscriber",
        "email",
        "currency",
        "default_source",
        "coupon",
        "balance",
        "business_vat_id",
    )
    list_filter = (CustomerHasSourceListFilter, CustomerSubscriptionStatusListFilter)
    search_fields = ("email", "description")
    inlines = (SubscriptionInline,)


@admin.register(models.Dispute)
class DisputeAdmin(StripeModelAdmin):
    list_display = ("reason", "status", "amount", "currency", "is_charge_refundable")
    list_filter = ("is_charge_refundable", "reason", "status")

    def has_add_permission(self, request):
        return False


@admin.register(models.Event)
class EventAdmin(StripeModelAdmin):
    list_display = ("type", "request_id")
    list_filter = ("type", "created")
    search_fields = ("request_id",)

    def has_add_permission(self, request):
        return False


@admin.register(models.FileUpload)
class FileUploadAdmin(StripeModelAdmin):
    list_display = ("purpose", "size", "type")
    list_filter = ("purpose", "type")
    search_fields = ("filename",)


@admin.register(models.PaymentIntent)
class PaymentIntentAdmin(StripeModelAdmin):
    list_display = (
        "id",
        "customer",
        "amount",
        "currency",
        "description",
        "amount_capturable",
        "amount_received",
        "receipt_email",
    )
    search_fields = ("customer__id", "invoice__id")


@admin.register(models.SetupIntent)
class SetupIntentAdmin(StripeModelAdmin):
    list_display = (
        "id",
        "created",
        "customer",
        "description",
        "on_behalf_of",
        "payment_method",
        "payment_method_types",
        "status",
    )
    list_filter = ("status",)
    search_fields = ("customer__id", "status")


@admin.register(models.Invoice)
class InvoiceAdmin(StripeModelAdmin):
    list_display = (
        "customer",
        "number",
        "paid",
        "forgiven",
        "closed",
        "period_start",
        "period_end",
        "subtotal",
        "tax",
        "tax_percent",
        "total",
    )
    list_filter = (
        "paid",
        "forgiven",
        "closed",
        "attempted",
        "created",
        "due_date",
        "period_start",
        "period_end",
    )
    search_fields = ("customer__id", "number", "receipt_number")
    inlines = (InvoiceItemInline,)


@admin.register(models.Plan)
class PlanAdmin(StripeModelAdmin):
    radio_fields = {"interval": admin.HORIZONTAL}

    def save_model(self, request, obj, form, change):
        """Update or create objects using our custom methods that sync with Stripe."""
        if change:
            obj.update_name()
        else:
            models.Plan.get_or_create(**form.cleaned_data)

    def get_readonly_fields(self, request, obj=None):
        """Return extra readonly_fields."""
        readonly_fields = super().get_readonly_fields(request, obj)

        if obj:
            readonly_fields += (
                "amount",
                "currency",
                "interval",
                "interval_count",
                "trial_period_days",
            )

        return readonly_fields


@admin.register(models.Product)
class ProductAdmin(StripeModelAdmin):
    list_display = ("name", "type", "active", "url", "statement_descriptor")
    list_filter = ("type", "active", "shippable")
    search_fields = ("name", "statement_descriptor")


@admin.register(models.Refund)
class RefundAdmin(StripeModelAdmin):
    list_display = (
        "amount",
        "currency",
        "charge",
        "reason",
        "status",
        "failure_reason",
    )
    list_filter = ("reason", "status")
    search_fields = ("receipt_number",)


@admin.register(models.Source)
class SourceAdmin(StripeModelAdmin):
    list_display = ("customer", "type", "status", "amount", "currency", "usage", "flow")
    list_filter = ("type", "status", "usage", "flow")


@admin.register(models.PaymentMethod)
class PaymentMethodAdmin(StripeModelAdmin):
    list_display = ("customer", "billing_details")
    list_filter = ("customer",)


@admin.register(models.Subscription)
class SubscriptionAdmin(StripeModelAdmin):
    list_display = ("customer", "status")
    list_filter = ("status", "cancel_at_period_end")

    inlines = (SubscriptionItemInline,)

    def cancel_subscription(self, request, queryset):
        """Cancel a subscription."""
        for subscription in queryset:
            subscription.cancel()

    cancel_subscription.short_description = "Cancel selected subscriptions"

    actions = (cancel_subscription,)


@admin.register(models.TaxRate)
class TaxRateAdmin(StripeModelAdmin):
    list_display = ("active", "display_name", "inclusive", "jurisdiction", "percentage")
    list_filter = ("active", "inclusive", "jurisdiction")


@admin.register(models.Transfer)
class TransferAdmin(StripeModelAdmin):
    list_display = ("amount", "description")



import stripe
from django.db import models, transaction
from stripe.error import InvalidRequestError

from .. import enums
from .. import settings as djstripe_settings
from ..exceptions import StripeObjectManipulationException
from ..fields import (
    JSONField,
    StripeCurrencyCodeField,
    StripeDecimalCurrencyAmountField,
    StripeEnumField,
)
from .base import StripeModel, logger
from .core import Customer


class DjstripePaymentMethod(models.Model):
    """
    An internal model that abstracts the legacy Card and BankAccount
    objects with Source objects.

    Contains two fields: `id` and `type`:
    - `id` is the id of the Stripe object.
    - `type` can be `card`, `bank_account` or `source`.
    """

    id = models.CharField(max_length=255, primary_key=True)
    type = models.CharField(max_length=12, db_index=True)

    @classmethod
    def from_stripe_object(cls, data):
        source_type = data["object"]
        model = cls._model_for_type(source_type)

        with transaction.atomic():
            model.sync_from_stripe_data(data)
            instance, _ = cls.objects.get_or_create(
                id=data["id"], defaults={"type": source_type}
            )

        return instance

    @classmethod
    def _get_or_create_source(cls, data, source_type):
        try:
            model = cls._model_for_type(source_type)
            model._get_or_create_from_stripe_object(data)
        except ValueError as e:
            # This may happen if we have source types we don't know about.
            # Let's not make dj-stripe entirely unusable if that happens.
            logger.warning("Could not sync source of type %r: %s", source_type, e)

        return cls.objects.get_or_create(id=data["id"], defaults={"type": source_type})

    @classmethod
    def _model_for_type(cls, type):
        if type == "card":
            return Card
        elif type == "source":
            return Source
        elif type == "bank_account":
            return BankAccount

        raise ValueError("Unknown source type: {}".format(type))

    @property
    def object_model(self):
        return self._model_for_type(self.type)

    def resolve(self):
        return self.object_model.objects.get(id=self.id)


class LegacySourceMixin:
    """
    Mixin for functionality shared between the legacy Card & BankAccount sources
    """

    @classmethod
    def _get_customer_from_kwargs(cls, **kwargs):
        if "customer" not in kwargs or not isinstance(kwargs["customer"], Customer):
            raise StripeObjectManipulationException(
                "{}s must be manipulated through a Customer. "
                "Pass a Customer object into this call.".format(cls.__name__)
            )

        customer = kwargs["customer"]
        del kwargs["customer"]

        return customer, kwargs

    @classmethod
    def _api_create(cls, api_key=djstripe_settings.STRIPE_SECRET_KEY, **kwargs):
        # OVERRIDING the parent version of this function
        # Cards & Bank Accounts must be manipulated through a customer or account.
        # TODO: When managed accounts are supported, this method needs to
        #     check if either a customer or account is supplied to determine
        #     the correct object to use.

        customer, clean_kwargs = cls._get_customer_from_kwargs(**kwargs)

        return customer.api_retrieve().sources.create(api_key=api_key, **clean_kwargs)

    @classmethod
    def api_list(cls, api_key=djstripe_settings.STRIPE_SECRET_KEY, **kwargs):
        # OVERRIDING the parent version of this function
        # Cards & Bank Accounts must be manipulated through a customer or account.
        # TODO: When managed accounts are supported, this method needs to
        #     check if either a customer or account is supplied to determine
        #     the correct object to use.

        customer, clean_kwargs = cls._get_customer_from_kwargs(**kwargs)

        return (
            customer.api_retrieve(api_key=api_key)
            .sources.list(object=cls.stripe_class.OBJECT_NAME, **clean_kwargs)
            .auto_paging_iter()
        )

    def get_stripe_dashboard_url(self):
        return self.customer.get_stripe_dashboard_url()

    def remove(self):
        """
        Removes a legacy source from this customer's account.
        """

        # First, wipe default source on all customers that use this card.
        Customer.objects.filter(default_source=self.id).update(default_source=None)

        try:
            self._api_delete()
        except InvalidRequestError as exc:
            if "No such source:" in str(exc) or "No such customer:" in str(exc):
                # The exception was thrown because the stripe customer or card
                # was already deleted on the stripe side, ignore the exception
                pass
            else:
                # The exception was raised for another reason, re-raise it
                raise

        self.delete()

    def api_retrieve(self, api_key=None, stripe_account=None):
        # OVERRIDING the parent version of this function
        # Cards & Banks Accounts must be manipulated through a customer or account.
        # TODO: When managed accounts are supported, this method needs to check if
        # either a customer or account is supplied to determine the
        # correct object to use.
        api_key = api_key or self.default_api_key
        customer = self.customer.api_retrieve(
            api_key=api_key, stripe_account=stripe_account
        )

        # If the customer is deleted, the sources attribute will be absent.
        # eg. {"id": "cus_XXXXXXXX", "deleted": True}
        if "sources" not in customer:
            # We fake a native stripe InvalidRequestError so that it's caught
            # like an invalid ID error.
            raise InvalidRequestError("No such source: %s" % (self.id), "id")

        # This will retrieve the source using the account ID where the customer resides,
        # so we don't have to pass `stripe_account`.
        return customer.sources.retrieve(self.id, expand=self.expand_fields)


class BankAccount(LegacySourceMixin, StripeModel):
    stripe_class = stripe.BankAccount

    account = models.ForeignKey(
        "Account",
        on_delete=models.PROTECT,
        null=True,
        blank=True,
        related_name="bank_account",
        help_text="The account the charge was made on behalf of. Null here indicates "
        "that this value was never set.",
    )
    account_holder_name = models.TextField(
        max_length=5000,
        default="",
        blank=True,
        help_text="The name of the person or business that owns the bank account.",
    )
    account_holder_type = StripeEnumField(
        enum=enums.BankAccountHolderType,
        help_text="The type of entity that holds the account.",
    )
    bank_name = models.CharField(
        max_length=255,
        help_text="Name of the bank associated with the routing number "
        "(e.g., `WELLS FARGO`).",
    )
    country = models.CharField(
        max_length=2,
        help_text="Two-letter ISO code representing the country the bank account "
        "is located in.",
    )
    currency = StripeCurrencyCodeField()
    customer = models.ForeignKey(
        "Customer", on_delete=models.SET_NULL, null=True, related_name="bank_account"
    )
    default_for_currency = models.NullBooleanField(
        help_text="Whether this external account is the default account for "
        "its currency."
    )
    fingerprint = models.CharField(
        max_length=16,
        help_text=(
            "Uniquely identifies this particular bank account. "
            "You can use this attribute to check whether two bank accounts are "
            "the same."
        ),
    )
    last4 = models.CharField(max_length=4)
    routing_number = models.CharField(
        max_length=255, help_text="The routing transit number for the bank account."
    )
    status = StripeEnumField(enum=enums.BankAccountStatus)


class Card(LegacySourceMixin, StripeModel):
    """
    You can store multiple cards on a customer in order to charge the customer later.

    This is a legacy model which only applies to the "v2" Stripe API (eg. Checkout.js).
    You should strive to use the Stripe "v3" API (eg. Stripe Elements).
    Also see: https://stripe.com/docs/stripe-js/elements/migrating
    When using Elements, you will not be using Card objects. Instead, you will use
    Source objects.
    A Source object of type "card" is equivalent to a Card object. However, Card
    objects cannot be converted into Source objects by Stripe at this time.

    Stripe documentation: https://stripe.com/docs/api/python#cards
    """

    stripe_class = stripe.Card

    address_city = models.TextField(
        max_length=5000,
        blank=True,
        default="",
        help_text="City/District/Suburb/Town/Village.",
    )
    address_country = models.TextField(
        max_length=5000, blank=True, default="", help_text="Billing address country."
    )
    address_line1 = models.TextField(
        max_length=5000,
        blank=True,
        default="",
        help_text="Street address/PO Box/Company name.",
    )
    address_line1_check = StripeEnumField(
        enum=enums.CardCheckResult,
        blank=True,
        default="",
        help_text="If `address_line1` was provided, results of the check.",
    )
    address_line2 = models.TextField(
        max_length=5000,
        blank=True,
        default="",
        help_text="Apartment/Suite/Unit/Building.",
    )
    address_state = models.TextField(
        max_length=5000,
        blank=True,
        default="",
        help_text="State/County/Province/Region.",
    )
    address_zip = models.TextField(
        max_length=5000, blank=True, default="", help_text="ZIP or postal code."
    )
    address_zip_check = StripeEnumField(
        enum=enums.CardCheckResult,
        blank=True,
        default="",
        help_text="If `address_zip` was provided, results of the check.",
    )
    brand = StripeEnumField(enum=enums.CardBrand, help_text="Card brand.")
    country = models.CharField(
        max_length=2,
        default="",
        blank=True,
        help_text="Two-letter ISO code representing the country of the card.",
    )
    customer = models.ForeignKey(
        "Customer", on_delete=models.SET_NULL, null=True, related_name="legacy_cards"
    )
    cvc_check = StripeEnumField(
        enum=enums.CardCheckResult,
        default="",
        blank=True,
        help_text="If a CVC was provided, results of the check.",
    )
    dynamic_last4 = models.CharField(
        max_length=4,
        default="",
        blank=True,
        help_text="(For tokenized numbers only.) The last four digits of the device "
        "account number.",
    )
    exp_month = models.IntegerField(help_text="Card expiration month.")
    exp_year = models.IntegerField(help_text="Card expiration year.")
    fingerprint = models.CharField(
        default="",
        blank=True,
        max_length=16,
        help_text="Uniquely identifies this particular card number.",
    )
    funding = StripeEnumField(
        enum=enums.CardFundingType, help_text="Card funding type."
    )
    last4 = models.CharField(max_length=4, help_text="Last four digits of Card number.")
    name = models.TextField(
        max_length=5000, default="", blank=True, help_text="Cardholder name."
    )
    tokenization_method = StripeEnumField(
        enum=enums.CardTokenizationMethod,
        default="",
        blank=True,
        help_text="If the card number is tokenized, this is the method that was used.",
    )

    def str_parts(self):
        return [
            "brand={brand}".format(brand=self.brand),
            "last4={last4}".format(last4=self.last4),
            "exp_month={exp_month}".format(exp_month=self.exp_month),
            "exp_year={exp_year}".format(exp_year=self.exp_year),
        ] + super().str_parts()

    @classmethod
    def create_token(
        cls,
        number,
        exp_month,
        exp_year,
        cvc,
        api_key=djstripe_settings.STRIPE_SECRET_KEY,
        **kwargs
    ):
        """
        Creates a single use token that wraps the details of a credit card.
        This token can be used in place of a credit card dictionary with any API method.
        These tokens can only be used once: by creating a new charge object,
        or attaching them to a customer.
        (Source: https://stripe.com/docs/api/python#create_card_token)

        :param number: The card number without any separators (no spaces)
        :type number: str
        :param exp_month: The card's expiration month. (two digits)
        :type exp_month: int
        :param exp_year: The card's expiration year. (four digits)
        :type exp_year: int
        :param cvc: Card security code.
        :type cvc: str
        :param api_key:
        :type api_key: str
        :rtype: stripe.Token
        """

        card = {
            "number": number,
            "exp_month": exp_month,
            "exp_year": exp_year,
            "cvc": cvc,
        }
        card.update(kwargs)

        return stripe.Token.create(api_key=api_key, card=card)


class Source(StripeModel):
    """
    Stripe documentation: https://stripe.com/docs/api#sources
    """

    amount = StripeDecimalCurrencyAmountField(
        null=True,
        blank=True,
        help_text=(
            "Amount (as decimal) associated with the source. "
            "This is the amount for which the source will be chargeable once ready. "
            "Required for `single_use` sources."
        ),
    )
    client_secret = models.CharField(
        max_length=255,
        help_text=(
            "The client secret of the source. "
            "Used for client-side retrieval using a publishable key."
        ),
    )
    currency = StripeCurrencyCodeField(default="", blank=True)
    flow = StripeEnumField(
        enum=enums.SourceFlow, help_text="The authentication flow of the source."
    )
    owner = JSONField(
        help_text=(
            "Information about the owner of the payment instrument that may be "
            "used or required by particular source types."
        )
    )
    statement_descriptor = models.CharField(
        max_length=255,
        default="",
        blank=True,
        help_text="Extra information about a source. This will appear on your "
        "customer's statement every time you charge the source.",
    )
    status = StripeEnumField(
        enum=enums.SourceStatus,
        help_text="The status of the source. Only `chargeable` sources can be used "
        "to create a charge.",
    )
    type = StripeEnumField(enum=enums.SourceType, help_text="The type of the source.")
    usage = StripeEnumField(
        enum=enums.SourceUsage,
        help_text="Whether this source should be reusable or not. "
        "Some source types may or may not be reusable by construction, "
        "while other may leave the option at creation.",
    )

    # Flows
    code_verification = JSONField(
        null=True,
        blank=True,
        help_text="Information related to the code verification flow. "
        "Present if the source is authenticated by a verification code "
        "(`flow` is `code_verification`).",
    )
    receiver = JSONField(
        null=True,
        blank=True,
        help_text="Information related to the receiver flow. "
        "Present if the source is a receiver (`flow` is `receiver`).",
    )
    redirect = JSONField(
        null=True,
        blank=True,
        help_text="Information related to the redirect flow. "
        "Present if the source is authenticated by a redirect (`flow` is `redirect`).",
    )

    source_data = JSONField(help_text="The data corresponding to the source type.")

    customer = models.ForeignKey(
        "Customer",
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        related_name="sources",
    )

    stripe_class = stripe.Source
    stripe_dashboard_item_name = "sources"

    @classmethod
    def _manipulate_stripe_object_hook(cls, data):
        # The source_data dict is an alias of all the source types
        data["source_data"] = data[data["type"]]
        return data

    def _attach_objects_hook(self, cls, data):
        customer = cls._stripe_object_to_customer(target_cls=Customer, data=data)
        if customer:
            self.customer = customer
        else:
            self.customer = None

    def detach(self):
        """
        Detach the source from its customer.

        :return:
        :rtype: bool
        """

        # First, wipe default source on all customers that use this.
        Customer.objects.filter(default_source=self.id).update(default_source=None)

        try:
            # TODO - we could use the return value of sync_from_stripe_data
            #  or call its internals - self._sync/_attach_objects_hook etc here
            #  to update `self` at this point?
            self.sync_from_stripe_data(self.api_retrieve().detach())
            return True
        except (InvalidRequestError, NotImplementedError):
            # The source was already detached. Resyncing.
            # NotImplementedError is an artifact of stripe-python<2.0
            # https://github.com/stripe/stripe-python/issues/376
            self.sync_from_stripe_data(self.api_retrieve())
            return False


class PaymentMethod(StripeModel):
    """
    Stripe documentation: https://stripe.com/docs/api#payment_methods
    """

    billing_details = JSONField(
        help_text=(
            "Billing information associated with the PaymentMethod that may be used or "
            "required by particular types of payment methods."
        )
    )
    card = JSONField(
        help_text="If this is a card PaymentMethod, this hash contains details "
        "about the card."
    )
    card_present = JSONField(
        null=True,
        blank=True,
        help_text="If this is an card_present PaymentMethod, this hash contains "
        "details about the Card Present payment method.",
    )
    customer = models.ForeignKey(
        "Customer",
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        related_name="payment_methods",
        help_text="Customer to which this PaymentMethod is saved."
        "This will not be set when the PaymentMethod has not been saved to a Customer.",
    )
    type = models.CharField(
        max_length=255,
        blank=True,
        help_text="The type of the PaymentMethod. An additional hash is included "
        "on the PaymentMethod with a name matching this value. It contains additional "
        "information specific to the PaymentMethod type.",
    )

    stripe_class = stripe.PaymentMethod

    def _attach_objects_hook(self, cls, data):
        customer = cls._stripe_object_to_customer(target_cls=Customer, data=data)
        if customer:
            self.customer = customer
        else:
            self.customer = None

    @classmethod
    def attach(
        cls, payment_method, customer, api_key=djstripe_settings.STRIPE_SECRET_KEY
    ):
        """
        Attach a payment method to a customer
        :param payment_method:
        :type payment_method: str, PaymentMethod
        :param customer:
        :type customer: Union[str, Customer]
        :param api_key:
        :type api_key: str
        :return:
        :rtype: PaymentMethod
        """

        if isinstance(payment_method, StripeModel):
            payment_method = payment_method.id

        if isinstance(customer, StripeModel):
            customer = customer.id

        extra_kwargs = {}
        if not isinstance(payment_method, stripe.PaymentMethod):
            # send api_key if we're not passing in a Stripe object
            # avoids "Received unknown parameter: api_key" since api uses the
            # key cached in the Stripe object
            extra_kwargs = {"api_key": api_key}

        stripe_payment_method = stripe.PaymentMethod.attach(
            payment_method, customer=customer, **extra_kwargs
        )
        return cls.sync_from_stripe_data(stripe_payment_method)

    def detach(self):
        """
        Detach the payment method from its customer.

        :return: Returns true if the payment method was newly detached, \
                 false if it was already detached
        :rtype: bool
        """
        # Find customers that use this
        customers = Customer.objects.filter(default_payment_method=self).all()
        changed = True

        # special handling is needed for legacy "card"-type PaymentMethods,
        # since detaching them deletes them within Stripe.
        # see https://github.com/dj-stripe/dj-stripe/pull/967
        is_legacy_card = self.id.startswith("card_")

        try:
            self.sync_from_stripe_data(self.api_retrieve().detach())

            # resync customer to update .default_payment_method and
            # .invoice_settings.default_payment_method
            for customer in customers:
                Customer.sync_from_stripe_data(customer.api_retrieve())

        except (InvalidRequestError,):
            # The source was already detached. Resyncing.

            if self.pk and not is_legacy_card:
                self.sync_from_stripe_data(self.api_retrieve())
            changed = False

        if self.pk:
            if is_legacy_card:
                self.delete()
            else:
                self.refresh_from_db()

        return changed



"""
dj-stripe model managers
"""
import decimal

from django.db import models


class StripeModelManager(models.Manager):
    """Manager used in StripeModel."""

    pass


class SubscriptionManager(models.Manager):
    """Manager used in models.Subscription."""

    def started_during(self, year, month):
        """Return Subscriptions not in trial status between a certain time range."""
        return self.exclude(status="trialing").filter(
            start__year=year, start__month=month
        )

    def active(self):
        """Return active Subscriptions."""
        return self.filter(status="active")

    def canceled(self):
        """Return canceled Subscriptions."""
        return self.filter(status="canceled")

    def canceled_during(self, year, month):
        """Return Subscriptions canceled during a certain time range."""
        return self.canceled().filter(canceled_at__year=year, canceled_at__month=month)

    def started_plan_summary_for(self, year, month):
        """Return started_during Subscriptions with plan counts annotated."""
        return (
            self.started_during(year, month)
            .values("plan")
            .order_by()
            .annotate(count=models.Count("plan"))
        )

    def active_plan_summary(self):
        """Return active Subscriptions with plan counts annotated."""
        return (
            self.active().values("plan").order_by().annotate(count=models.Count("plan"))
        )

    def canceled_plan_summary_for(self, year, month):
        """
        Return Subscriptions canceled within a time range with plan counts annotated.
        """
        return (
            self.canceled_during(year, month)
            .values("plan")
            .order_by()
            .annotate(count=models.Count("plan"))
        )

    def churn(self):
        """Return number of canceled Subscriptions divided by active Subscriptions."""
        canceled = self.canceled().count()
        active = self.active().count()
        return decimal.Decimal(str(canceled)) / decimal.Decimal(str(active))


class TransferManager(models.Manager):
    """Manager used by models.Transfer."""

    def during(self, year, month):
        """Return Transfers between a certain time range."""
        return self.filter(created__year=year, created__month=month)

    def paid_totals_for(self, year, month):
        """
        Return paid Transfers during a certain year, month with total amounts annotated.
        """
        return self.during(year, month).aggregate(total_amount=models.Sum("amount"))


class ChargeManager(models.Manager):
    """Manager used by models.Charge."""

    def during(self, year, month):
        """Return Charges between a certain time range based on `created`."""
        return self.filter(created__year=year, created__month=month)

    def paid_totals_for(self, year, month):
        """
        Return paid Charges during a certain year, month with total amount,
        fee and refunded annotated.
        """
        return (
            self.during(year, month)
            .filter(paid=True)
            .aggregate(
                total_amount=models.Sum("amount"),
                total_refunded=models.Sum("amount_refunded"),
            )
        )



def example():
    import djstripe.models
    import djstripe.settings
    import stripe

    # stripe API return value is a dict-like object
    stripe_data = stripe.Product.create(
        api_key=djstripe.settings.STRIPE_SECRET_KEY,
        name="Monthly membership base fee",
        type="service",
    )

    # sync_from_stripe_data to save it to the database,
    # and recursively update any referenced objects
    djstripe_obj = djstripe.models.Product.sync_from_stripe_data(stripe_data)

    return djstripe_obj



"""
Utility functions related to the djstripe app.
"""

import datetime

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser
from django.core.exceptions import ImproperlyConfigured
from django.db.models.query import QuerySet
from django.utils import timezone

ANONYMOUS_USER_ERROR_MSG = (
    "dj-stripe's payment checking mechanisms require the user "
    "be authenticated before use. Please use django.contrib.auth's "
    "login_required decorator or a LoginRequiredMixin. "
    "Please read the warning at "
    "http://dj-stripe.readthedocs.org/en/latest/usage.html#ongoing-subscriptions."
)


def fix_django_headers(meta):
    """
    Fix this nonsensical API:
    https://docs.djangoproject.com/en/1.11/ref/request-response/
    https://code.djangoproject.com/ticket/20147
    """
    ret = {}
    for k, v in meta.items():
        if k.startswith("HTTP_"):
            k = k[len("HTTP_") :]
        elif k not in ("CONTENT_LENGTH", "CONTENT_TYPE"):
            # Skip CGI garbage
            continue

        ret[k.lower().replace("_", "-")] = v

    return ret


def subscriber_has_active_subscription(subscriber, plan=None):
    """
    Helper function to check if a subscriber has an active subscription.

    Throws improperlyConfigured if the subscriber is an instance of AUTH_USER_MODEL
    and get_user_model().is_anonymous == True.

    Activate subscription rules (or):
        * customer has active subscription

    If the subscriber is an instance of AUTH_USER_MODEL, active subscription rules (or):
        * customer has active subscription
        * user.is_superuser
        * user.is_staff

    :param subscriber: The subscriber for which to check for an active subscription.
    :type subscriber: dj-stripe subscriber
    :param plan: The plan for which to check for an active subscription.
        If plan is None and there exists only one subscription, this method will
        check if that subscription is active. Calling this method with no plan and
        multiple subscriptions will throw an exception.
    :type plan: Plan or string (plan ID)

    """
    if isinstance(subscriber, AnonymousUser):
        raise ImproperlyConfigured(ANONYMOUS_USER_ERROR_MSG)

    if isinstance(subscriber, get_user_model()):
        if subscriber.is_superuser or subscriber.is_staff:
            return True
    from .models import Customer

    customer, created = Customer.get_or_create(subscriber)
    if created or not customer.has_active_subscription(plan):
        return False
    return True


def get_supported_currency_choices(api_key):
    """
    Pull a stripe account's supported currencies and returns a choices tuple of those
    supported currencies.

    :param api_key: The api key associated with the account from which to pull data.
    :type api_key: str
    """
    import stripe

    stripe.api_key = api_key

    account = stripe.Account.retrieve()
    supported_payment_currencies = stripe.CountrySpec.retrieve(account["country"])[
        "supported_payment_currencies"
    ]

    return [(currency, currency.upper()) for currency in supported_payment_currencies]


def clear_expired_idempotency_keys():
    from .models import IdempotencyKey

    threshold = timezone.now() - datetime.timedelta(hours=24)
    IdempotencyKey.objects.filter(created__lt=threshold).delete()


def convert_tstamp(response):
    """
    Convert a Stripe API timestamp response (unix epoch) to a native datetime.

    :rtype: datetime
    """
    if response is None:
        # Allow passing None to convert_tstamp()
        return response

    # Overrides the set timezone to UTC - I think...
    tz = timezone.utc if settings.USE_TZ else None

    return datetime.datetime.fromtimestamp(response, tz)


# TODO: Finish this.
CURRENCY_SIGILS = {"CAD": "$", "EUR": "€", "GBP": "£", "USD": "$"}


def get_friendly_currency_amount(amount, currency):
    currency = currency.upper()
    sigil = CURRENCY_SIGILS.get(currency, "")
    return "{sigil}{amount:.2f} {currency}".format(
        sigil=sigil, amount=amount, currency=currency
    )


class QuerySetMock(QuerySet):
    """
    A mocked QuerySet class that does not handle updates.
    Used by UpcomingInvoice.invoiceitems.
    """

    @classmethod
    def from_iterable(cls, model, iterable):
        instance = cls(model)
        instance._result_cache = list(iterable)
        instance._prefetch_done = True
        return instance

    def _clone(self):
        return self.__class__.from_iterable(self.model, self._result_cache)

    def update(self):
        return 0

    def delete(self):
        return 0



from unittest.mock import patch

from django.test import TestCase
from docs.usage.examples import manually_syncing_with_stripe

import djstripe.models


class CheckExamplesTest(TestCase):
    """
    Sanity check example code
    """

    @patch("stripe.Product.create", autospec=True)
    def test_manually_sync_data_with_stripe(self, product_create_mock):
        example_product_data = {
            "id": "example_product",
            "object": "product",
            "active": True,
            "name": "Monthly membership base fee",
            "type": "service",
        }

        product_create_mock.return_value = example_product_data

        manually_syncing_with_stripe.example()

        self.assertEqual(
            djstripe.models.Product.objects.last().id, example_product_data["id"]
        )



from django.core.management.base import BaseCommand

from ...utils import clear_expired_idempotency_keys


class Command(BaseCommand):
    help = "Deleted expired Stripe idempotency keys."

    def handle(self, *args, **options):
        clear_expired_idempotency_keys()



# Generated by Django 2.0.6 on 2018-06-27 08:21

from django.db import migrations, models

import djstripe.fields


class Migration(migrations.Migration):

    dependencies = [("djstripe", "0001_initial")]

    operations = [
        migrations.AlterField(
            model_name="account",
            name="business_name",
            field=models.CharField(
                help_text="The publicly visible name of the business",
                max_length=255,
                null=True,
                blank=True,
            ),
        ),
        migrations.AlterField(
            model_name="account",
            name="support_url",
            field=models.CharField(
                help_text="A publicly shareable URL that provides support for this account",
                max_length=200,
                null=True,
                blank=True,
            ),
        ),
        migrations.AlterField(
            model_name="charge",
            name="receipt_number",
            field=models.CharField(
                help_text="The transaction number that appears on email receipts sent for this charge.",
                max_length=14,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="product",
            name="deactivate_on",
            field=djstripe.fields.JSONField(
                help_text="An array of connect application identifiers that cannot purchase this product. Only applicable to products of `type=good`.",
                null=True,
                blank=True,
            ),
        ),
        migrations.AlterField(
            model_name="product",
            name="images",
            field=djstripe.fields.JSONField(
                help_text="A list of up to 8 URLs of images for this product, meant to be displayable to the customer. Only applicable to products of `type=good`.",
                null=True,
                blank=True,
            ),
        ),
    ]



"""
init_customers command.
"""
from django.core.management.base import BaseCommand

from ...models import Customer
from ...settings import get_subscriber_model


class Command(BaseCommand):
    """Create customer objects for existing subscribers that don't have one."""

    help = "Create customer objects for existing subscribers that don't have one"

    def handle(self, *args, **options):
        """
        Create Customer objects for Subscribers without Customer objects associated.
        """
        for subscriber in get_subscriber_model().objects.filter(
            djstripe_customers=None
        ):
            # use get_or_create in case of race conditions on large subscriber bases
            Customer.get_or_create(subscriber=subscriber)
            print("Created subscriber for {0}".format(subscriber.email))
